/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package ring;

import org.objectweb.dream.aaa.AgentServer;
import org.objectweb.dream.aaa.agent.AgentId;
import org.objectweb.dream.aaa.agent.Notification;
import org.objectweb.dream.aaa.agent.PushNotification;
import org.objectweb.fractal.api.Component;

/**
 * Ring benchmark. Deploy Agents organized on a virtual ring. A notification is
 * forwarded around the ring. Statistics are printed periodicaly.
 */
public final class Main
{

  private Main()
  {
  }

  private static int     payload = -1;
  private static String  a3servers;
  private static String  serverName;
  private static short[] regIDs;

  /**
   * Entry point. Create a local server using the given <code>servername0</code>
   * then deploy a {@link RingAgent }on each given repositoryID.
   * 
   * @param args <code>[-p payload] a3serversfile servername [repID0]...</code>
   * @throws Exception if an error occurs.
   */
  public static void main(String[] args) throws Exception
  {
    parseArgs(args);
    System.out.println("Deploy a ring of " + regIDs.length + " agents");
    Component s0 = AgentServer.createAgentServer(a3servers, serverName);
    PushNotification pn0 = AgentServer.getPushNotificationItf(s0);
    short r0 = AgentServer.getDefaultRepositoryId(s0);

    // create agents and organize the virtual ring.
    RingAgent[] agents = new RingAgent[regIDs.length];
    for (int i = 0; i < regIDs.length; i++)
    {
      agents[i] = new RingAgent(r0, regIDs[i]);
      if (i > 0)
      {
        agents[i - 1].nextAgent = agents[i].getId();
      }
    }
    agents[regIDs.length - 1].nextAgent = agents[0].getId();

    // deploy agents.
    for (int i = 0; i < agents.length; i++)
    {
      agents[i].deploy(pn0);
    }

    // send notification around the ring
    Notification not;
    if (payload < 0)
    {
      not = new Notification();
    }
    else
    {
      not = new RingNotif(payload);
    }
    pn0.pushNotification(null, agents[0].getId(), not);
    System.out.println("Agents deployed, notification sent.");

    // print statistics
    AgentId statID = agents[0].getId();
    while (true)
    {
      Thread.sleep(2000);
      pn0.pushNotification(null, statID, new PrintStatRequestNotif());
    }

  }

  private static void parseArgs(String[] args)
  {
    if (args.length < 3)
    {
      usage();
    }
    int nextArg = 0;
    if (args[0].equals("-p"))
    {
      try
      {
        payload = Integer.parseInt(args[1]);
      }
      catch (NumberFormatException e)
      {
        System.err.println("Invalid payload : " + e.getMessage());
        usage();
      }
      nextArg += 2;
    }
    a3servers = args[nextArg++];
    serverName = args[nextArg++];
    regIDs = new short[args.length - nextArg];
    for (int i = 0; i < regIDs.length; i++)
    {
      try
      {
        regIDs[i] = Short.parseShort(args[nextArg + i]);
      }
      catch (NumberFormatException e)
      {
        System.err.println("Invalid repositoryID : " + e.getMessage());
        usage();
      }
    }
  }

  private static void usage()
  {
    System.err.println("Usage : ring.Main [-p payload] a3serversfile "
        + "servername repID0 [repID1] ...");
    System.err.println("Run a ring benchmark. Deploy Agents organized on a "
        + "virtual ring. ");
    System.err.println("A notification is forwarded around the ring. "
        + "Statistics are printed periodicaly.");
    System.err.println();
    System.err.println("    -p payload       the message payload in byte "
        + "(no payload by default).");
    System.err.println("    a3serversfile    URL of the a3servers.xml "
        + "configuration file.");
    System.err.println("    servername       the name of the local server to "
        + "start.");
    System.err.println("    repID0           the ID of the local repository.");
    System.err.println("    repID1..n        the ID of the remote "
        + "repositories on which agents will be deployed.");
    System.exit(1);
  }
}