/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): Vivien Quema
 */

/*
 * Parts of this code are copyright:
 * 
 * The contents of this file are subject to the Joram Public License, as defined
 * by the file JORAM_LICENSE.TXT
 * 
 * You may not use this file except in compliance with the License. You may
 * obtain a copy of the License on the Objectweb web site (www.objectweb.org).
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License.
 * 
 * The Original Code is Joram, including the java packages fr.dyade.aaa.agent,
 * fr.dyade.aaa.util, fr.dyade.aaa.ip, fr.dyade.aaa.mom, and fr.dyade.aaa.joram,
 * released May 24, 2000.
 * 
 * The Initial Developer of the Original Code is Dyade. The Original Code and
 * portions created by Dyade are Copyright Bull and Copyright INRIA. All Rights
 * Reserved.
 * 
 * The present code contributor is INRIA Rhone-Alpes.
 */

package org.objectweb.dream.aaa.agent;

import java.io.IOException;
import java.io.Serializable;
import java.lang.reflect.Field;
import java.lang.reflect.Method;

import org.objectweb.dream.PushException;
import org.objectweb.dream.util.Error;
import org.objectweb.dream.util.NullLogger;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * /** The <code>Agent</code> class represents the basic component in our
 * model. <i>agents </i> are "reactive" objects which behave according to "event ->
 * reaction"model: an event embodies a significant state change which one or
 * many agents may react to.
 * <p>
 * Class <code>Agent</code> defines the generic interface and the common
 * behavior for all agents; every agent is an object of a class deriving from
 * class Agent. Agents are the elementary programming and execution entities;
 * they only communicate using notifications through the message bus, and are
 * controlled by the execution engine.
 * <p>
 * The reactive behavior is implemented by function member React, which defines
 * the reaction of the agent when receiving a notification; this function member
 * is called by the execution engine.
 * <p>
 * Agents are persistent objects, and the Agent class realizes a
 * "swap-in/swap-outPushItf" mechanism which allows loading (or finding) in main
 * memory the agents to activate, and unloading the agents idle since a while.
 * <p>
 * <hr>
 * Agents must be created in two steps:
 * <ul>
 * <li>locally creating the object in memory (via constructor),
 * <li>configure it (for example via get/set methods),
 * <li>then deploy it .
 * </ul>
 * <p>
 * The following code would then create a simple agent and deploy it:
 * <p>
 * <blockquote>
 * 
 * <pre>
 * Agent ag = new Agent();
 * ag.deploy();
 * </pre>
 * 
 * </blockquote>
 * <p>
 */
public abstract class Agent implements Serializable
{

  private static final String        NULL_NAME = "";

  /** Symbolic name of the agent */
  protected transient String         name;

  /** A logger that can be used bye the agent */
  protected transient Logger         logger    = NullLogger.NULL_LOGGER;

  /**
   * Global unique identifier of the agent. Each agent is identified by a unique
   * identifier allowing the agent to be found. The identifiers format is
   * detailed in <a href="AgentId.html">AgentId </a> class.
   */
  protected transient AgentId        id;

  /**
   * Some agents must be loaded at any time, this can be enforced by this member
   * variable. If <code>true</code> agent is pinned in memory.
   */
  protected transient boolean        fixed;

  /**
   * The send notification interface
   */
  private transient PushNotification pushNotificationItf;

  /**
   * Determines if the currently <code>Agent</code> has already been deployed.
   */
  protected boolean                  deployed  = false;

  private void writeObject(java.io.ObjectOutputStream out) throws IOException
  {
    out.writeShort(id.from);
    out.writeShort(id.to);
    out.writeInt(id.stamp);
    out.writeUTF(name);
    out.writeBoolean(fixed);
  }

  private void readObject(java.io.ObjectInputStream in) throws IOException,
      ClassNotFoundException
  {
    id = new AgentId(in.readShort(), in.readShort(), in.readInt());
    if ((name = in.readUTF()).equals(NULL_NAME))
    {
      name = NULL_NAME;
    }
    fixed = in.readBoolean();
    logger = NullLogger.NULL_LOGGER;
  }

  /**
   * Set the execution environnement for this agent
   * 
   * @param pushNotificationItf the fractal interface reference used to send
   *          notifications
   * @param logger a logger
   */
  protected final void setEnvironnement(PushNotification pushNotificationItf,
      Logger logger)
  {
    this.logger = logger;
    this.pushNotificationItf = pushNotificationItf;
  }

  /**
   * Allocates a new Agent object. This constructor has the same effect as
   * <code>Agent(local, local, null, false)</code>.
   * 
   * @param local Identication of the local agent server
   * @see Agent#Agent(short, short, String, boolean)
   */
  public Agent(short local)
  {
    this(local, local, null, false);
  }

  /**
   * Allocates a new Agent object. This constructor has the same effect as
   * <code>Agent(local, to, null, false)</code>.
   * 
   * @param local Identication of the local agent server
   * @param to Identication of target agent server
   * @see Agent#Agent(short, short, String, boolean)
   */
  public Agent(short local, short to)
  {
    this(local, to, null, false);
  }

  /**
   * Allocates a new Agent object. This constructor has the same effect as
   * <code>Agent(to, name, false)</code>.
   * 
   * @param local Identication of the local agent server
   * @param to Identication of target agent server
   * @param name symbolic name
   * @see Agent#Agent(short, short, String, boolean)
   */
  public Agent(short local, short to, String name)
  {
    this(local, to, name, false);
  }

  /**
   * Allocates a new Agent object. This constructor has the same effect as
   * <code>Agent(to, null, fixed)</code>.
   * 
   * @param local Identication of the local agent server
   * @param to Identication of target agent server
   * @param fixed if <code>true</code> agent is pinned in memory
   * @see Agent#Agent(short, short, String, boolean)
   */
  public Agent(short local, short to, boolean fixed)
  {
    this(local, to, null, fixed);
  }

  /**
   * Allocates a new Agent object. The resulting object <b>is not an agent </b>;
   * before it can react to a notification you must deploy it.
   * 
   * @param local Identication of the local agent server
   * @param to Identication of target agent server
   * @param name symbolic name
   * @param fixed if <code>true</code> agent is pinned in memory
   * @see #deploy(PushNotification)
   */
  public Agent(short local, short to, String name, boolean fixed)
  {
    AgentId id = new AgentId(local, to);
    initState(name, fixed, id);
  }

  /**
   * Constructor used to build Well Known Services agents.
   * <p>
   * System agents are created from the <code>agent</code> package. WKS agents
   * are similar to system agents, except that they may be defined in separate
   * packages, and they do not necessarily exist on all agent servers. Their
   * creation is controlled from the configuration file of the agent server.
   * <p>
   * This constructor takes the agent id as a parameter instead of building it.
   * Since the constructor has been made public, the consistency of agent ids
   * allocation must be enforced. This is done by the constructor checking that
   * the id stamp is comprised in the <code>AgentId.MIN_WKS_ID_STAMP</code>-
   * <code>AgentId.MaxWKSIdStamp</code> interval.
   * 
   * @param to Identication of local (and target) agent server
   * @param name symbolic name
   * @param fixed if <code>true</code> agent is pinned in memory
   * @param stamp well known stamp
   */
  public Agent(short to, String name, boolean fixed, int stamp)
  {
    if (stamp < AgentId.MIN_WKS_ID_STAMP || stamp > AgentId.MAX_WKS_ID_STAMP)
    {
      logger.log(BasicLevel.ERROR,
          "Creation, well known service stamp out of range: " + stamp);
    }
    // TO DO : call method on agentReactor
    AgentId id = new AgentId(to, to, stamp);
    initState(name, fixed, id);
  }

  /**
   * Constructor used to build "system" agents like <code>FactoryReact</code>.
   * System agents are created from the <code>agent</code> package. This
   * constructor takes the agent id as a parameter instead of building it.
   * 
   * @param name symbolic name
   * @param fixed if <code>true</code> agent is pinned in memory
   * @param id the id of the agent
   */
  Agent(String name, boolean fixed, AgentId id)
  {
    initState(name, fixed, id);
  }

  private void initState(String name, boolean fixed, AgentId id)
  {
    if (name == null)
    {
      this.name = NULL_NAME;
    }
    else
    {
      this.name = name;
    }
    this.fixed = fixed;
    this.id = id;
  }

  /**
   * Returns if the currently <code>Agent</code> has already been deployed.
   * 
   * @return if the currently <code>Agent</code> has already been deployed.
   */
  public boolean isDeployed()
  {
    return deployed;
  }

  /**
   * Deploys a new <i>agent </i>. It works by sending a notification to a
   * special agent, of class Factory, running on the target agent server. The
   * notification asks for a remote creation of the agent. This solution
   * presents the advantage of reusing the standard communication mechanisms of
   * the agent machine.
   * <p>
   * The whole process involves then the following steps:
   * <ul>
   * <li>serializing the object state,
   * <li>building an <code>AgentCreateRequest</code> notification with the
   * resulting bytes stream,
   * <li>sending it to the target Factory agent.
   * </ul>
   * In reaction, the factory agent builds the agent in the target server from
   * the serialized image, and saves it into operational storage.
   * 
   * @param pushNotificationItf the interface used to send the notification to
   *          the factory.
   * @exception IOException unspecialized exception
   */
  public final void deploy(PushNotification pushNotificationItf)
      throws IOException
  {
    deploy(null, pushNotificationItf);
  }

  /**
   * Deploys a new <i>agent </i>. It works as <a href="#deploy()">deploy() </a>
   * method above; after the agent creation, the Factory agent sends an
   * <code>AgentCreateReply</code> notification.
   * 
   * @param reply agent to reply to
   * @param pushNotificationItf the interface used to send the notification to
   *          the factory.
   * @exception IOException unspecialized exception
   */
  public final void deploy(AgentId reply, PushNotification pushNotificationItf)
      throws IOException
  {
    if ((id == null) || id.isNullId())
    {
      logger.log(BasicLevel.ERROR, " can't deploy agent [" + name
          + "]: id is null");
      throw new IOException("Can't deploy agent [" + name + "]: id is null");
    }
    if (deployed)
    {
      logger.log(BasicLevel.ERROR, " can't deploy agent [" + name
          + "]: already deployed");
      throw new IOException("Can't deploy agent [" + name
          + "]: already deployed");
    }

    /** ***** Warning de Dede non pris en compte pour la suite : ******** */
    //  If we use sendTo agent's method the from field is the agent id, and
    // on reception the from node (from.to) can be false.
    /** *** */
    try
    {
      pushNotificationItf.pushNotification(null, AgentId.factoryId(id.to),
          new AgentCreateRequest(this, reply));
    }
    catch (PushException e)
    {
      Error.error("Unexpected error while sending notification", logger, e);
    }

    deployed = true;

    if (logger.isLoggable(BasicLevel.DEBUG))
    {
      logger.log(BasicLevel.DEBUG, "Agent" + id + " [" + name + "] deployed");
    }
  }

  /**
   * Protected method which can be used by sub classes to deploy an agent. This
   * is usefull when an agent deploy an other agent. <br>
   * This method must be used during a reaction.
   * 
   * @param agent the agent to deploy
   * @throws IOException unspecialized exception
   */
  protected final void deploy(Agent agent) throws IOException
  {
    agent.deploy(pushNotificationItf);
  }

  /**
   * Protected method which can be used by sub classes to deploy an agent. This
   * is usefull when an agent deploy an other agent. <br>
   * This method must be used during a reaction.
   * 
   * @param agent the agent to deploy
   * @param reply agent to reply to
   * @throws IOException unspecialized exception
   * @see Agent#deploy(AgentId, PushNotification)
   */
  protected final void deploy(Agent agent, AgentId reply) throws IOException
  {
    agent.deploy(reply, pushNotificationItf);
  }

  /**
   * returns the name of this agent.
   * 
   * @return the name of this agent.
   */
  public String getName()
  {
    if (name == null)
    {
      return getClass().getName() + id.toString();
    }
    else
    {
      return name;
    }
  }

  /**
   * Returns a string representation of this agent, including the agent's class,
   * name, global identication, and fixed property.
   * 
   * @return A string representation of this agent.
   */
  public String toString()
  {
    return "(" + super.toString() + ",name=" + name + ",id=" + id.toString()
        + ",fixed=" + fixed + ")";
  }

  /**
   * Returns the global unique identifier of the agent. Each agent is identified
   * by a unique identifier allowing the agent to be found. The identifiers
   * format is detailed in <a href="AgentId.html">AgentId </a> class.
   * 
   * @return the global unique identifier of the agent.
   */
  public final AgentId getId()
  {
    return id;
  }

  /**
   * Tests if the agent is pinned in memory.
   * 
   * @return true if this agent is a pinned in memory; false otherwise.
   * @see #fixed
   */
  public final boolean isFixed()
  {
    return fixed;
  }

  /**
   * Gives this agent an opportunity to initialize after having been deployed,
   * and each time it is loaded into memory.
   * <p>
   * This function is first called by the factory agent, just after it deploys
   * the agent.
   * <p>
   * This function is used by agents with a <code>fixed</code> field set to
   * <code>true</code> to initialize their transient variables, as it is
   * called each time the agent server is restarted.
   * <p>
   * This function is not declared <code>final</code> so that derived classes
   * may change their reload policy. The implementation of this method provided
   * by the <code>Agent</code> class does nothing.
   * 
   * @param firstTime true when first called by the factory
   * @exception Exception unspecialized exception
   */
  protected void initialize(boolean firstTime) throws Exception
  {
    // Get the logging monitor from current server MonologLoggerFactory
    if (logger.isLoggable(BasicLevel.DEBUG))
    {
      logger.log(BasicLevel.DEBUG, "Agent" + id + " [" + name
          + (firstTime ? "] , first initialized" : "] , initialized"));
    }
  }

  /**
   * This method sends a notification to the agent which id is given in
   * parameter.
   * <p>
   * Be careful to never use this method outside of an agent reaction, otherwise
   * its behavior is undefined. If you needs to send notification outside of a
   * reaction you must use <code>Channel.sendTo</code> method.
   * 
   * @param to the unique id. of destination <code>Agent</code>.
   * @param not the notification to send.
   */
  protected final void sendTo(AgentId to, Notification not)
  {
    try
    {
      pushNotificationItf.pushNotification(id, to, not);
    }
    catch (PushException e)
    {
      Error.error("Unexpected error while sending notification", logger, e);
    }
  }

  /**
   * Permits this agent to destroy it. If necessary, its method should be
   * overloaded to work properly.
   * 
   * @param pushNotificationItf the interface used to send the notification to
   *          the factory.
   */
  public void delete(PushNotification pushNotificationItf)
  {
    delete(null, pushNotificationItf);
  }

  /**
   * Permits this agent to destroy it. If necessary, its method should be
   * overloaded to work properly.
   * 
   * @param pushNotificationItf the interface used to send the notification to
   *          the factory.
   * @param agent Id of agent to notify.
   */
  public void delete(AgentId agent, PushNotification pushNotificationItf)
  {
    if (deployed)
    {
      try
      {
        pushNotificationItf.pushNotification(getId(), AgentId.factoryId(id.to),
            new AgentDeleteRequest(agent));
      }
      catch (PushException e)
      {
        Error.error("Unexpected error while sending notification", logger, e);
      }
    }
  }

  /**
   * Protected method which can be used by sub classes to delete an other agent.
   * <br>
   * This method must be used during a reaction.
   * 
   * @param agent the agent to delete
   */
  protected final void delete(Agent agent)
  {
    agent.delete(pushNotificationItf);
  }

  /**
   * Protected method which can be used by sub classes to delete an other agent.
   * <br>
   * This method must be used during a reaction.
   * 
   * @param agent the agent to delete
   * @param reply agent to reply to
   */
  protected final void delete(Agent agent, AgentId reply)
  {
    agent.delete(reply, pushNotificationItf);
  }

  /**
   * Defines the reaction of the agent when receiving a notification. This
   * member function implements the common reactive behavior of an agent, it is
   * called by the execution engine (see <a href="Engine.html">Engine </a>
   * class).
   * <ul>
   * <li>Configuration/Reconfiguration: <code>SetField</code>,
   * <code>DupRequest</code>,<code>DeleteNot</code>
   * <li>Monitoring: <code>SubscribeNot</code>,<code>GetStatusNot</code>,
   * <li>Error: <code>UnknownAgent</code>,<code>UnknownNotification</code>,
   * <code>ExceptionNotification</code>
   * </ul>
   * <p>
   * If there is no corresponding reaction, the agent send an
   * <code>UnknownNotification</code> notification to the sender.
   * 
   * @param from agent sending notification
   * @param not notification to react to
   * @exception Exception unspecialized exception
   */
  public void react(AgentId from, Notification not) throws Exception
  {
    if (not instanceof SetField)
    {
      setField((SetField) not);
      if (((SetField) not).getReply() != null)
      {
        sendTo(((SetField) not).getReply(), new SetFieldAck());
      }
    }
    else if (not instanceof DupRequest)
    {
      AgentId newId = dup((DupRequest) not, null);
      sendTo(from, new DupReply(newId));
    }
    else if (not instanceof DeleteNot)
    {
      delete(((DeleteNot) not).getReply(), pushNotificationItf);
      //  } else if (not instanceof SubscribeNot) {
      // doReact(from,(SubscribeNot) not);
      //    } else if (not instanceof GetStatusNot) {
      // doReact(from,(GetStatusNot) not);
    }
    else if (not instanceof UnknownAgent)
    {
      //  if (AgentServer.MONITOR_AGENT) {
      //    doReact((UnknownAgent) not);
      //  }
    }
    else if (not instanceof UnknownNotification)
    {
      // ...
    }
    else
    {
      sendTo(from, new UnknownNotification(id, not));
    }
  }

  /**
   * set field
   * 
   * @param not the notification
   * @throws Exception an exception
   */
  protected void setField(SetField not) throws Exception
  {
    // It seems it's not necessary to protect "id" and "fixed" fields
    // but I can't explain why!!!
    //  if ((not.name.equals("id")) || (not.name.equals("fixed")))
    //     throw new IllegalAccessException("Can't change Field:"
    // + not.name);

    // test if a specific setAttribute method exist
    String methName;
    byte tmp[] = not.getName().getBytes();
    if ((tmp[0] >= 'a') && (tmp[0] <= 'z'))
    {
      // we must set to upcase the first letter of the attribute name
      // the sepecific method convention name is "setAtt", att is the
      // name of the attribute
      tmp[0] = (byte) (tmp[0] - ('a' - 'A'));

      methName = new String("set" + new String(tmp));
    }
    else
    {
      // make the method name
      methName = new String("set" + not.getName());
    }

    // Search the agent classe for calling the specific method
    Class agClass = ((Object) this).getClass();

    // Search the parameter class of the specific method
    // the classe of the attribute to be set must be an object.
    // the specific method always have one parameter
    Class paramClass[] = new Class[1];
    Object paramValue[] = new Object[1];
    paramClass[0] = not.getValue().getClass();
    paramValue[0] = not.getValue();

    try
    {
      // get the method for invoke it
      Method setMeth = agClass.getMethod(methName, paramClass);
      // the result of the method is always void
      setMeth.invoke(this, paramValue);
    }
    catch (NoSuchMethodException e)
    {
      // the getmethod failled, we try to use generic setfield
      Field field = getClass().getField(not.getName());
      field.set(this, not.getValue());
    }
    catch (SecurityException e)
    {
      // the getmethod failled, we try to use generic setfield
      Field field = getClass().getField(not.getName());
      field.set(this, not.getValue());
    }
  }

  /**
   * Method to react to <code>DupRequest</code> notifications. It deploys a
   * copy of current agent on <code>not.to</code> agent server
   * 
   * @param not notification to react to
   * @param reply agent to sending creation acknowledge
   * @return the agentId of the created agent.
   * @throws Exception unspecialized exception
   */
  protected AgentId dup(DupRequest not, AgentId reply) throws Exception
  {
    AgentId oldId = id;
    AgentId newId = new AgentId(not.getTo());

    try
    {
      id = newId;
      deployed = false;
      deploy(reply, pushNotificationItf);
      id = oldId;
    }
    catch (IOException exc)
    {
      id = oldId;
      throw exc;
    }

    return newId;
  }

  /**
   * Called to inform this agent that it is garbaged and that it should free any
   * active ressources that it has allocated. A subclass of <code>Agent</code>
   * should override this method if it has any operation that it wants to
   * perform before it is garbaged. For example, an agent with threads (a
   * ProxyAgent for example) would use the initialize method to create the
   * threads and the <code>agentFinalize</code> method to stop them. The
   * implementation of this method provided by the <code>Agent</code> class
   * does nothing.
   */
  public void agentFinalize()
  {
  }
}