/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): Vivien Quema
 */

package org.objectweb.dream.aaa.agent;

import java.util.Hashtable;
import java.util.Map;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.PushException;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

/**
 * The repository component keep informations to retreive Agent.
 */
public class RepositoryImpl extends AbstractComponent
    implements
      Repository,
      PushNotification,
      RepositoryAttributeController
{

  /**
   * The name of the client interface binded to the agent factory.
   */
  public static final String FACTORY_REACT_ITF_NAME         = "factoryReact";

  /**
   * The name of the client interface used be agents to send notifications
   */
  public static final String OUT_PUSH_NOTIFICATION_ITF_NAME = "out-push-notification";

  /**
   * The {@link ComponentReact}server interface reference to which this
   * component is binded
   */
  protected ComponentReact   factoryReactItf;

  /**
   * The interface used by agents to send notifications. This interface is not
   * used directly by this component, but it will be passed to agent before its
   * reaction.
   */
  protected PushNotification pushNotificationItf;

  /** Map of agents */
  private Map                agents;

  /** the repository identifier */
  private short              repositoryId;

  /**
   * Default constructor, create the repository
   */
  public RepositoryImpl()
  {
    agents = new Hashtable();
  }

  // ---------------------------------------------------------------------------
  // Implementation of Repository interface
  // ---------------------------------------------------------------------------

  /**
   * @see Repository#getRepositoryId()
   */
  public short getRepositoryId()
  {
    return repositoryId;
  }

  /**
   * @see Repository#addAgent(AgentId, Agent)
   */
  public void addAgent(AgentId agentId, Agent agent)
  {
    agents.put(agentId, agent);
    // set agent environnemt
    Logger agentLogger;
    String agentLoggerName = (agent.getName() == null) ? agent.getId()
        .toString() : agent.getName();
    if (logger instanceof LoggerFactory)
    {
      agentLogger = ((LoggerFactory) logger).getLogger(agentLoggerName);
    }
    else
    {
      agentLogger = Monolog.monologFactory.getLogger(agentLoggerName);
    }
    agent.setEnvironnement(pushNotificationItf, agentLogger);
    if (logger.isLoggable(BasicLevel.DEBUG))
    {
      logger.log(BasicLevel.DEBUG, "Agent added : " + agent);
    }
  }

  /**
   * @see Repository#removeAgent(AgentId)
   */
  public void removeAgent(AgentId agentId) throws UnknownAgentException
  {
    Object old = agents.remove(agentId);
    if (logger.isLoggable(BasicLevel.DEBUG))
    {
      logger.log(BasicLevel.DEBUG, "Agent removed : " + old);
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of PushNotification interface
  // ---------------------------------------------------------------------------

  /**
   * @see PushNotification#pushNotification(AgentId, AgentId, Notification)
   */
  public void pushNotification(AgentId from, AgentId to, Notification not)
      throws PushException
  {
    try
    {
      Agent agent = (Agent) agents.get(to);
      if (agent == null)
      {
        if (to.getStamp() == AgentId.FACTORY_ID_STAMP)
        {
          // notification for factory
          factoryReactItf.react(from, not);
        }
        else
        {
          throw new UnknownAgentException("Unknown AgentId : " + to);
        }
      }
      else
      {
        // make it to react.
        if (logger.isLoggable(BasicLevel.DEBUG))
        {
          logger.log(BasicLevel.DEBUG, "react agent=" + agent + ", from="
              + from + ", to=" + to + ", notification " + not);
        }
        agent.react(from, not);
      }
    }
    catch (Exception e)
    {
      throw new PushException("An error occurs during reaction", e);
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of RepositoryAttributeController interfaces
  // ---------------------------------------------------------------------------

  /**
   * @see RepositoryAttributeController#setRepositoryId(short)
   */
  public void setRepositoryId(short repositoryId)
  {
    this.repositoryId = repositoryId;
  }

  // ---------------------------------------------------------------------------
  // Implementation of BindingController interfaces
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{FACTORY_REACT_ITF_NAME, OUT_PUSH_NOTIFICATION_ITF_NAME};
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public synchronized void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    if (clientItfName.equals(FACTORY_REACT_ITF_NAME))
    {
      factoryReactItf = (ComponentReact) serverItf;
    }
    else if (clientItfName.equals(OUT_PUSH_NOTIFICATION_ITF_NAME))
    {
      pushNotificationItf = (PushNotification) serverItf;
    }
    super.bindFc(clientItfName, serverItf);
  }
}