/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): Vivien Quema
 */

package org.objectweb.dream.aaa.components;

import java.util.Map;

import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.PushPushDreamComponent;
import org.objectweb.dream.aaa.agent.UnknownServerException;
import org.objectweb.dream.aaa.components.conf.AgentServerConfiguration;
import org.objectweb.dream.aaa.components.conf.RepositoryDesc;
import org.objectweb.dream.message.ChunkAlreadyExistsException;
import org.objectweb.dream.message.ExtensibleMessage;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.manager.MessageManager;
import org.objectweb.dream.protocol.ArrowChunk;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Transformer adding a {@link ArrowChunk }chunk to message used by causality
 * protocol.
 */
public class CausalityTransformerImpl extends PushPushDreamComponent
    implements
      CausalityTransformerAttributeController
{

  private MessageManager           messageManagerItf;
  private AgentServerConfiguration agentServerConfigurationItf;

  private String                   arrowChunkName;
  private short                    localProcessId;

  // ---------------------------------------------------------------------------
  // Implementation of Push interface
  // ---------------------------------------------------------------------------

  /**
   * @see Push#push(Message, Map)
   */
  public void push(Message message, Map context) throws PushException
  {
    ArrowChunk arrowChunk = (ArrowChunk) message.getChunk(arrowChunkName);
    if (arrowChunk == null)
    {
      if (message instanceof ExtensibleMessage)
      {
        arrowChunk = (ArrowChunk) messageManagerItf
            .createChunk(ArrowChunk.TYPE);
        try
        {
          ((ExtensibleMessage) arrowChunk).addChunk(arrowChunkName,
              ArrowChunk.TYPE, arrowChunk);
        }
        catch (ChunkAlreadyExistsException e)
        {
          // cannot happend
        }
      }
      else
      {
        logger
            .log(
                BasicLevel.ERROR,
                "Unable to find arrow chunk in the unextensible message. The message is dropped");
        return;
      }
    }

    AgentChunk agentChunk = (AgentChunk) message
        .getChunk(AgentChunk.DEFAULT_NAME);
    if (agentChunk == null)
    {
      logger.log(BasicLevel.ERROR, "Unable to find agent chunk named "
          + AgentChunk.DEFAULT_NAME + ". The message is dropped");
      return;
    }

    arrowChunk.setProcessIdFrom(localProcessId);
    short index = index(agentChunk.getAgentTo().getTo());
    if (index < 0)
    {
      messageManagerItf.deleteMessage(message);
      throw new PushException(new UnknownServerException(
          "unknonw destination repository : " + agentChunk.getAgentTo().getTo()
              + ". message dropped"));

    }
    arrowChunk.setProcessIdTo(index);
    outPushItf.push(message, null);
  }

  // ---------------------------------------------------------------------------
  // Implementation of CausalityTransformerAttributeController interface
  // ---------------------------------------------------------------------------

  /**
   * @see CausalityTransformerAttributeController#getArrowChunkName()
   */
  public String getArrowChunkName()
  {
    return arrowChunkName;
  }

  /**
   * @see CausalityTransformerAttributeController#setArrowChunkName(String)
   */
  public void setArrowChunkName(String name)
  {
    arrowChunkName = name;
  }

  /**
   * @see CausalityTransformerAttributeController#getLocalDomainIndex()
   */
  public short getLocalDomainIndex()
  {
    return localProcessId;
  }

  /**
   * @see CausalityTransformerAttributeController#setLocalDomainIndex(short)
   */
  public void setLocalDomainIndex(short index)
  {
    localProcessId = index;
  }

  // ---------------------------------------------------------------------------
  // Implementation of BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{Push.OUT_PUSH_ITF_NAME, MessageManager.ITF_NAME};
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public synchronized void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }
    else if (clientItfName.equals(AgentServerConfiguration.DEFAULT_NAME))
    {
      agentServerConfigurationItf = (AgentServerConfiguration) serverItf;
    }
  }

  // ---------------------------------------------------------------------------
  // Utility method
  // ---------------------------------------------------------------------------

  private short index(short regId) throws PushException
  {
    RepositoryDesc reg = agentServerConfigurationItf.getRepositoryDesc(regId);
    if (reg == null)
    {
      return -1;
    }
    return reg.getServer().getAccessNetwork().getDomainIndex();
  }

}