/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): Vivien Quema
 */

package org.objectweb.dream.aaa.components.conf;

import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.julia.ChainedNoSuchInterfaceException;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;

/**
 * {@link A3ServerXmlParser }implementation component using SAX parser.
 */
public class A3ServerXmlContentHandler extends DefaultHandler
    implements
      A3ServerXmlParser,
      BindingController
{

  /** The syntaxic name for config element */
  public static final String CONFIG_ELEMENT_NAME     = "config";
  /** The syntaxic name for domain element */
  public static final String DOMAIN_ELEMENT_NAME     = "domain";
  /** The syntaxic name for property element */
  public static final String PROPERTY_ELEMENT_NAME   = "property";
  /** The syntaxic name for server element */
  public static final String SERVER_ELEMENT_NAME     = "server";
  /** The syntaxic name for network element */
  public static final String NETWORK_ELEMENT_NAME    = "network";
  /** The syntaxic name for the repository element */
  public static final String REPOSITORY_ELEMENT_NAME = "repository";
  /** The syntaxic name for id attribute */
  public static final String ID_ATTR_NAME            = "id";
  /** The syntaxic name for name attribute */
  public static final String NAME_ATTR_NAME          = "name";
  /** The syntaxic name for value attribute */
  public static final String VALUE_ATTR_NAME         = "value";
  /** Then syntaxic namee for codec attribute */
  public static final String CODEC_ATTR_NAME         = "codec";
  /** The syntaxic name for ADL name attribute */
  public static final String ADL_ATTR_NAME           = "adl";
  /** The syntaxic name for hostname attribute */
  public static final String HOSTNAME_ATTR_NAME      = "hostname";
  /** The syntaxic name for port attribute */
  public static final String PORT_ATTR_NAME          = "port";
  /** The syntaxic name for domain name attribute */
  public static final String DOMAIN_ATTR_NAME        = "domain";
  /** The syntaxic name for default attribute */
  public static final String DEFAULT_ATTR_NAME       = "default";

  /** The name of the {@link XMLReader}client interface */
  public static final String XML_READER_ITF_NAME     = "xml-reader";

  private XMLReader          parser;

  private Locator            locator;

  private List               serverList;
  private Map                domainMap;
  private ServerDesc         curServer;
  private DomainDesc         curDomain;
  private NetworkDesc        curNetwork;
  private Set                repositoryIdSet;

  /**
   * @see A3ServerXmlParser#parse(String)
   */
  public List parse(String url) throws Exception
  {
    serverList = new ArrayList();
    domainMap = new Hashtable();
    repositoryIdSet = new HashSet();
    curServer = null;
    curDomain = null;
    curNetwork = null;
    parser.parse(new InputSource(url));
    return serverList;
  }

  /**
   * @see org.xml.sax.ContentHandler#endElement(java.lang.String,
   *      java.lang.String, java.lang.String)
   */
  public void endElement(String uri, String localName, String qName)
      throws SAXException
  {
    if (localName.equals(SERVER_ELEMENT_NAME))
    {
      serverList.add(curServer);
      curServer = null;
    }
    else if (localName.equals(DOMAIN_ELEMENT_NAME))
    {
      curDomain = null;
    }
    else if (localName.equals(NETWORK_ELEMENT_NAME))
    {
      curNetwork = null;
    }
  }

  /**
   * @see org.xml.sax.ContentHandler#setDocumentLocator(org.xml.sax.Locator)
   */
  public void setDocumentLocator(Locator locator)
  {
    this.locator = locator;
  }

  /**
   * @see org.xml.sax.ContentHandler#startElement(java.lang.String,
   *      java.lang.String, java.lang.String, org.xml.sax.Attributes)
   */
  public void startElement(String uri, String localName, String qNmae,
      Attributes attributes) throws SAXException
  {
    if (localName.equals(CONFIG_ELEMENT_NAME))
    {
      // do nothing
    }
    else if (localName.equals(DOMAIN_ELEMENT_NAME))
    {
      String name = getAttribute(attributes, NAME_ATTR_NAME, "name of domain");
      String adlName = getAttribute(attributes, ADL_ATTR_NAME,
          "network default ADL ");
      String codec = getAttribute(attributes, CODEC_ATTR_NAME,
          "codec of domain");
      DomainDesc domain = new DomainDesc(name, codec, adlName);
      domainMap.put(name, domain);
      curDomain = domain;
    }
    else if (localName.equals(PROPERTY_ELEMENT_NAME))
    {
      String name = getAttribute(attributes, NAME_ATTR_NAME, "property name");
      String value = getAttribute(attributes, VALUE_ATTR_NAME, "property value");
      if (curDomain != null)
      {
        // domain property
        curDomain.addDefaultNetworkProperty(name, value);
      }
      else 
      {
        // network property
        curNetwork.addProperty(name, value);
      }
    }
    else if (localName.equals(SERVER_ELEMENT_NAME))
    {
      String name = getAttribute(attributes, NAME_ATTR_NAME, "name of server");
      String hostname = getAttribute(attributes, HOSTNAME_ATTR_NAME,
          "hostname of server");

      try
      {
        curServer = new ServerDesc(name, hostname);
      }
      catch (UnknownHostException e)
      {
        throw new SAXParseException(
            "Invalid hostname attribute for server element", locator, e);
      }
    }
    else if (localName.equals(REPOSITORY_ELEMENT_NAME))
    {
      String id = getAttribute(attributes, ID_ATTR_NAME, "id of repository");
      String name = getAttribute(attributes, NAME_ATTR_NAME,
          "name of repositoryrepositoryrepository");
      String adlName = getAttribute(attributes, ADL_ATTR_NAME,
          "engine ADL name");
      String defaultAttr = attributes.getValue(DEFAULT_ATTR_NAME);
      boolean isDefault = false;
      if (defaultAttr != null)
      {
        isDefault = Boolean.valueOf(defaultAttr).booleanValue();
      }
      try
      {
        short idVal = Short.parseShort(id);
        if (!repositoryIdSet.add(new Short(idVal)))
        {
          throw new SAXParseException("Duplicated repository identifer",
              locator);
        }
        RepositoryDesc reg = new RepositoryDesc(name, adlName, idVal, curServer);
        curServer.addRepository(reg, isDefault);
      }
      catch (NumberFormatException e)
      {
        throw new SAXParseException("Invalid id attribute for server element",
            locator, e);
      }
    }
    else if (localName.equals(NETWORK_ELEMENT_NAME))
    {
      String port = getAttribute(attributes, PORT_ATTR_NAME,
          "listenning port of network");
      String domainName = getAttribute(attributes, DOMAIN_ATTR_NAME,
          "domain name of network");
      DomainDesc domain = (DomainDesc) domainMap.get(domainName);
      if (domain == null)
      {
        throw new SAXParseException(
            "Invalid domain name for network element, unknown domain name",
            locator);
      }
      String adl = attributes.getValue(CODEC_ATTR_NAME);
      adl = (adl == null) ? domain.getDefaultNetworkADL() : adl;
      int portNum;
      try
      {
        portNum = Integer.parseInt(port);
      }
      catch (NumberFormatException e)
      {
        throw new SAXParseException(
            "Invalid port attribute for network element", locator, e);
      }
      NetworkDesc net = new NetworkDesc(curServer, domain, adl, portNum, domain
          .getDefaultNetworkProperties());
      curServer.addNetwork(net);
      domain.addNetwork(net);
      curNetwork = net;
    }
    else 
    {
      throw new SAXParseException("Unknown element : " + localName, locator);
    }

  }

  private String getAttribute(Attributes atts, String name, String errorMsg)
      throws SAXException
  {
    String s = atts.getValue(name);
    if (s == null)
    {
      throw new SAXParseException("Unable to find attribute \"" + name
          + "\" for " + errorMsg, locator);
    }
    return s;
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{XML_READER_ITF_NAME};
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#lookupFc(java.lang.String)
   */
  public Object lookupFc(String clientItfName) throws NoSuchInterfaceException
  {
    if (clientItfName.equals(XML_READER_ITF_NAME))
    {
      return parser;
    }
    else
    {
      throw new ChainedNoSuchInterfaceException(null, null,
          "No such interface : " + clientItfName);
    }
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(java.lang.String,
   *      java.lang.Object)
   */
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    if (clientItfName.equals(XML_READER_ITF_NAME))
    {
      parser = (XMLReader) serverItf;
    }
    else
    {
      throw new ChainedNoSuchInterfaceException(null, null,
          "No such interface : " + clientItfName);
    }
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#unbindFc(java.lang.String)
   */
  public void unbindFc(String clientItfName) throws NoSuchInterfaceException,
      IllegalBindingException, IllegalLifeCycleException
  {
    if (clientItfName.equals(XML_READER_ITF_NAME))
    {
      parser = null;
    }
    else
    {
      throw new ChainedNoSuchInterfaceException(null, null,
          "No such interface : " + clientItfName);
    }
  }

}