/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.aaa.components.conf;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.adl.Reconfiguration;
import org.objectweb.dream.adl.ReconfigurationFactory;
import org.objectweb.dream.message.codec.CodecManager;
import org.objectweb.dream.message.codec.CodecManagerException;
import org.objectweb.dream.util.Error;
import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.SuperController;
import org.objectweb.fractal.util.Fractal;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Agent server Attribute controller implemented like a component. It requires a
 * parser interface used to parse a3 configuration file.
 */
public class AgentServerConfigurationImpl extends AbstractComponent
    implements
      AgentServerInitialization,
      AgentServerConfiguration,
      BindingController
{

  /**
   * The name of the client interface bound to the component interface of the
   * Agent server composite
   */
  public static final String  AGENT_SERVER_COMPONENT_ITF_NAME = "agent-server-component";

  /** An array of every repository descriptions, indexed by their registry id */
  protected RepositoryDesc[]  repositories                    = null;
  /** The server description of the server instance, this component belong */
  protected ServerDesc        localServerDesc                 = null;

  protected String            serverName                      = null;
  protected String            a3ServerFileName;

  /** The parser server interface, which this component is binded. */
  protected A3ServerXmlParser parserItf;
  /** The {@link Component }interface of the agent server composite. */
  protected Component         agentServerComponentItf;
  /** The {@link CodecManager }of the agent server composite. */
  protected CodecManager      codecManagerItf;

  //----------------------------------------------------------------------------
  // Implementations of AgentServerInitialization interface
  //----------------------------------------------------------------------------

  /**
   * @see AgentServerInitialization#initialize(String, String)
   */
  public void initialize(String a3serversFile, String serverName)
      throws Exception
  {
    this.a3ServerFileName = a3serversFile;
    this.serverName = serverName;
    List serverList;
    try
    {
      serverList = parserItf.parse(a3ServerFileName);
    }
    catch (Exception e)
    {
      throw new Exception("Invalid a3server.xml file", e);
    }
    initRepositoriesArray(serverList);
    initThisServer();
    if (logger.isLoggable(BasicLevel.INFO))
    {
      logger.log(BasicLevel.INFO, "Server " + serverName + " initialized");
    }
  }

  //----------------------------------------------------------------------------
  // Implementations of AgentServerConfiguration interface
  //----------------------------------------------------------------------------

  /**
   * @see AgentServerConfiguration#getRepositoryDesc(String)
   */
  public RepositoryDesc getRepositoryDesc(String repositoryName)
  {
    if (localServerDesc == null)
    {
      throw new IllegalStateException("Unknown configuration");
    }
    List repositories = localServerDesc.getRepositories();
    Iterator iter = repositories.iterator();
    while (iter.hasNext())
    {
      RepositoryDesc repository = (RepositoryDesc) iter.next();
      if (repository.getName().equals(repositoryName))
      {
        return repository;
      }
    }
    return null;
  }

  /**
   * @see AgentServerConfiguration#getDefaultRepositoryDesc()
   */
  public RepositoryDesc getDefaultRepositoryDesc()
  {
    if (localServerDesc == null)
    {
      throw new IllegalStateException("Unknown configuration");
    }
    return localServerDesc.getDefaultRepository();
  }

  /**
   * @see AgentServerConfiguration#getNetworkDesc(String)
   */
  public NetworkDesc getNetworkDesc(String domainName)
  {
    if (localServerDesc == null)
    {
      throw new IllegalStateException("Unknown configuration");
    }
    Iterator iter = localServerDesc.getNetworks().iterator();
    while (iter.hasNext())
    {
      NetworkDesc network = (NetworkDesc) iter.next();
      if (network.getDomain().getName().equals(domainName))
      {
        return network;
      }
    }
    return null;
  }

  /**
   * @see AgentServerConfiguration#getRepositoryDesc(short)
   */
  public RepositoryDesc getRepositoryDesc(short repositoryId)
  {
    if (repositoryId >= repositories.length || repositoryId < 0)
    {
      return null;
    }
    return repositories[repositoryId];
  }

  /**
   * @see AgentServerConfiguration#getAccessibleRepositories(String)
   */
  public short[] getAccessibleRepositories(String messageConsumerName)
  {
    NetworkDesc networkDesc = getNetworkDesc(messageConsumerName);
    if (networkDesc != null)
    {
      List nets = networkDesc.getServer().getNetworks();
      if (nets.size() == 1)
      {
        // single network configuration
        return null;
      }
      else
      {
        // multiple networks configuration
        DomainDesc domain = networkDesc.getDomain();
        ServerDesc thisServerDesc = networkDesc.getServer();
        List accessRegs = new ArrayList();
        for (int i = 0; i < repositories.length; i++)
        {
          RepositoryDesc repository = repositories[i];
          if (repository == null)
          {
            continue;
          }
          ServerDesc server = repository.getServer();
          if (server == thisServerDesc)
          {
            continue;
          }
          if (server.getAccessNetwork().getDomain() == domain)
          {
            // this server is accessible through my domain
            // so i can access its Repositories
            accessRegs.add(repository);
          }
        }

        short[] accessibleRepositories = new short[accessRegs.size()];
        for (int i = 0; i < accessRegs.size(); i++)
        {
          accessibleRepositories[i] = ((RepositoryDesc) accessRegs.get(i))
              .getRepositoryID();
        }
        return accessibleRepositories;
      }
    }
    RepositoryDesc repositoryDesc = getRepositoryDesc(messageConsumerName);
    if (repositoryDesc != null)
    {
      short[] accessibleRepositories = new short[1];
      accessibleRepositories[0] = repositoryDesc.getRepositoryID();
      return accessibleRepositories;
    }
    throw new IllegalArgumentException("Unknown message consumer :"
        + messageConsumerName);
  }

  //----------------------------------------------------------------------------
  // Utility methods
  //----------------------------------------------------------------------------

  protected void initServer()
  {
  }

  /**
   * Initializes the <code>repositories</code> field of this class.
   * 
   * @param servers the list of servers
   * @throws Exception if the configuration of the servers is invalid.
   */
  protected void initRepositoriesArray(List servers) throws Exception
  {
    // add registries in registryMap
    Set repositorySet = new HashSet();
    // set of every servers
    Set serverSet = new HashSet(servers);
    short maxRegId = 0;
    Iterator iter = servers.iterator();
    while (iter.hasNext())
    {
      ServerDesc server = (ServerDesc) iter.next();
      List repositories = server.getRepositories();
      Iterator repositoryiter = repositories.iterator();
      while (repositoryiter.hasNext())
      {
        RepositoryDesc rep = (RepositoryDesc) repositoryiter.next();
        repositorySet.add(rep);
        if (rep.getRepositoryID() > maxRegId)
        {
          maxRegId = rep.getRepositoryID();
        }
      }
      if (server.getName().equals(serverName))
      {
        localServerDesc = server;
      }
    }
    repositories = new RepositoryDesc[maxRegId + 1];
    iter = repositorySet.iterator();
    while (iter.hasNext())
    {
      RepositoryDesc rep = (RepositoryDesc) iter.next();
      repositories[rep.getRepositoryID()] = rep;
    }
    if (localServerDesc == null)
    {
      throw new RuntimeException(serverName + " : no such sever name in "
          + a3ServerFileName);
    }

    // search a route for each server
    List currentLevelDomain = new ArrayList();
    for (int i = 0; i < localServerDesc.getNetworks().size(); i++)
    {
      DomainDesc d = ((NetworkDesc) localServerDesc.getNetworks().get(i))
          .getDomain();
      currentLevelDomain.add(d);
    }
    do
    {
      List nextLevelDomain = new ArrayList();
      Iterator currentLevelDomainIterator = currentLevelDomain.iterator();
      while (currentLevelDomainIterator.hasNext())
      {
        DomainDesc curDomain = (DomainDesc) currentLevelDomainIterator.next();

        // gateway of the curDomain,
        // if null, this server has a network for this gateway
        ServerDesc gateway = curDomain.getGateway();
        Iterator curDomainNetworksIterator = curDomain.getNetworks().iterator();
        while (curDomainNetworksIterator.hasNext())
        {
          NetworkDesc net = (NetworkDesc) curDomainNetworksIterator.next();
          ServerDesc server = net.getServer();
          serverSet.remove(server);
          if (server == localServerDesc)
          {
            server.setAccessNetwork((NetworkDesc) server.getNetworks().get(0));
            continue;
          }
          if (server.isVisited())
          {
            continue;
          }
          server.visite();
          for (int k = 0; k < server.getNetworks().size(); k++)
          {
            // for each network of each server in the curent domain ...
            NetworkDesc network = (NetworkDesc) server.getNetworks().get(k);
            if (network.getDomain() == curDomain)
            {
              // this network give access to the current domain
              if (gateway == null)
              {
                // this server can be accessed directly
                server.setAccessNetwork(network);
              }
              else
              {
                // this server is accessed though the gateway.
                server.setAccessNetwork(gateway.getAccessNetwork());
              }
            }
            else
            {
              // this network access to a new domain
              DomainDesc newDomain = network.getDomain();
              if (newDomain.getGateway() != null)
              {
                // domain already visited
                throw new RuntimeException("Cycle domain");
              }
              nextLevelDomain.add(newDomain);
              if (gateway == null)
              {
                // this new domain is accessible directly through the server
                newDomain.setGateway(server);
              }
              else
              {
                // this new domain is accessible through the current gateway
                newDomain.setGateway(gateway);
              }
            }
          } // end of 'for each network in the server'
        } // end of 'for each server in the domain'
      } // end of 'for each domain in the current level'
      currentLevelDomain = nextLevelDomain;
    }
    while (currentLevelDomain.size() != 0);

    if (!serverSet.isEmpty())
    {
      throw new Exception("Invalid domain graph, non connex graph");
    }

    // for each networks directly accessible throught local networks
    // set a unique index in its domain.
    Iterator localNetworksIterator = localServerDesc.getNetworks().iterator();
    while (localNetworksIterator.hasNext())
    {
      NetworkDesc localNetwork = (NetworkDesc) localNetworksIterator.next();
      Object[] networks = localNetwork.getDomain().getNetworks().toArray();
      Arrays.sort(networks);
      for (short i = 0; i < networks.length; i++)
      {
        ((NetworkDesc) networks[i]).setDomainIndex(i);
      }

    }
  }

  protected void initThisServer() throws ADLException, CodecManagerException,
      IllegalLifeCycleException
  {
    List registryList = localServerDesc.getRepositories();
    for (int i = 0; i < registryList.size(); i++)
    {
      initEngine((RepositoryDesc) registryList.get(i));
    }
    List networkList = localServerDesc.getNetworks();
    for (int i = 0; i < networkList.size(); i++)
    {
      initNetwork((NetworkDesc) networkList.get(i));
    }
  }

  protected void initNetwork(NetworkDesc networkDesc) throws ADLException,
      CodecManagerException, IllegalLifeCycleException
  {
    String domainName = networkDesc.getDomain().getName();
    String codecName = networkDesc.getDomain().getCodecName();
    String networkName = "Network-" + domainName;
    
    Map context = new HashMap();
    context.put("networkADL", networkDesc.getAdl());
    context.put("domainName", domainName);
    context.put("listenPort", "" + networkDesc.getPort());
    context.put("nbNetwork", "" + networkDesc.getDomain().getNetworks().size());
    context.put("localIndex", "" + networkDesc.getDomainIndex());
    context.putAll(networkDesc.getProperties());
    
    codecManagerItf.addCodec(codecName, context);
    Component codec = codecManagerItf.getCodec(codecName);
    Reconfiguration reconf = ReconfigurationFactory.getReconfiguration();

    context.put("legacy:./" + networkName + "/Codec", codec);
    reconf.addComponents(agentServerComponentItf,
        "org.objectweb.dream.aaa.components.conf.AddNetwork", context);

    // start network.
    try
    {
      Component[] subComponent = Fractal.getContentController(
          agentServerComponentItf).getFcSubComponents();
      for (int i = 0; i < subComponent.length; i++)
      {
        Component component = subComponent[i];
        if (networkName
            .equals(Fractal.getNameController(component).getFcName()))
        {
          Fractal.getLifeCycleController(component).startFc();
        }
      }
    }
    catch (NoSuchInterfaceException e)
    {
      Error.bug(logger, e);
    }
  }

  /**
   * Initializes an engine from the specified registry description.
   * 
   * @param repositoryDesc the registry description
   * @throws ADLException if the creation fails
   */
  protected void initEngine(RepositoryDesc repositoryDesc) throws ADLException,
      IllegalLifeCycleException
  {
    String engineName = "Engine-" + repositoryDesc.getName();
    Reconfiguration reconf = ReconfigurationFactory.getReconfiguration();
    Map context = new HashMap();
    context.put("engineADL", repositoryDesc.getEngineADLName());
    context.put("repositoryName", repositoryDesc.getName());
    context.put("repositoryId", "" + repositoryDesc.getRepositoryID());
    reconf.addComponents(agentServerComponentItf,
        "org.objectweb.dream.aaa.components.conf.AddEngine", context);

    // start engine.
    try
    {
      Component[] subComponent = Fractal.getContentController(
          agentServerComponentItf).getFcSubComponents();
      for (int i = 0; i < subComponent.length; i++)
      {
        Component component = subComponent[i];
        if (engineName.equals(Fractal.getNameController(component).getFcName()))
        {
          Fractal.getLifeCycleController(component).startFc();
        }
      }
    }
    catch (NoSuchInterfaceException e)
    {
      Error.bug(logger, e);
    }
  }

  protected String findPath(Component composite, Component subComponent)
  {
    String name;
    try
    {
      name = Fractal.getNameController(subComponent).getFcName();
    }
    catch (NoSuchInterfaceException e1)
    {
      return null;
    }
    SuperController sc;
    try
    {
      sc = Fractal.getSuperController(subComponent);
    }
    catch (NoSuchInterfaceException e)
    {
      return null;
    }

    Component[] superComponents = sc.getFcSuperComponents();
    for (int i = 0; i < superComponents.length; i++)
    {
      Component superComponent = superComponents[i];
      if (superComponent == composite)
      {
        return "./" + name;
      }
      String path = findPath(composite, superComponent);
      if (path != null)
      {
        return path + "/" + name;
      }
    }
    return null;
  }

  // ---------------------------------------------------------------------------
  // Implementation of BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{A3ServerXmlParser.ITF_NAME,
        AGENT_SERVER_COMPONENT_ITF_NAME, CodecManager.ITF_NAME};
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(A3ServerXmlParser.ITF_NAME))
    {
      parserItf = (A3ServerXmlParser) serverItf;
    }
    else if (clientItfName.equals(AGENT_SERVER_COMPONENT_ITF_NAME))
    {
      agentServerComponentItf = (Component) serverItf;
    }
    else if (clientItfName.equals(CodecManager.ITF_NAME))
    {
      codecManagerItf = (CodecManager) serverItf;
    }
  }
}