/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): Vivien Quema
 */

package org.objectweb.dream.channel;

import java.io.IOException;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.codec.MessageCodec;
import org.objectweb.dream.message.manager.MessageManager;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Generic sh ChannelIn implementation.
 */
public class GenericPushChannelInImpl extends AbstractComponent
    implements
      OpenedSocket
{

  /**
   * The name of the optional client interface used by this coponent to send an
   * acknowlegdment
   */
  public static final String SEND_ACK_OPT_ITF_NAME = "send-ack-opt";

  // ---------------------------------------------------------------------------
  // Client interfaces
  // ---------------------------------------------------------------------------

  protected Push             outPushItf;
  protected MessageManager   messageManagerItf;
  protected MessageCodec     messageCodecItf;
  protected SendByte         sendAckOptItf;

  // ---------------------------------------------------------------------------
  // Implementation of the OpenedSocket interface
  // ---------------------------------------------------------------------------

  /**
   * @see OpenedSocket#openedSocket(SocketState)
   */
  public void openedSocket(SocketState socket) throws IOException
  {
    Message message = messageCodecItf.decode(socket);

    logger.log(BasicLevel.DEBUG, "msg received");
    try
    {
      outPushItf.push(message, null);
    }
    catch (PushException e)
    {
      if (sendAckOptItf != null)
      {
        logger.log(BasicLevel.ERROR, "A push exception occurs. Send a nack", e);
        sendAckOptItf.sendByte(socket.getOutput(), 1);
      }
      else
      {
        logger.log(BasicLevel.ERROR, "A push exception occurs", e);
      }
      messageManagerItf.deleteMessage(message);
      return;
    }

    // then send the acknowledge.
    if (sendAckOptItf != null)
    {
      logger.log(BasicLevel.DEBUG, "send ack");
      sendAckOptItf.sendByte(socket.getOutput(), 0);
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{MessageManager.ITF_NAME, MessageCodec.ITF_NAME,
        Push.OUT_PUSH_ITF_NAME, SEND_ACK_OPT_ITF_NAME};
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(Push.OUT_PUSH_ITF_NAME))
    {
      outPushItf = (Push) serverItf;
    }
    else if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }
    else if (clientItfName.equals(MessageCodec.ITF_NAME))
    {
      messageCodecItf = (MessageCodec) serverItf;
    }
    else if (clientItfName.equals(SEND_ACK_OPT_ITF_NAME))
    {
      sendAckOptItf = (SendByte) serverItf;
    }
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#unbindFc(String)
   */
  public void unbindFc(String clientItfName) throws NoSuchInterfaceException,
      IllegalBindingException, IllegalLifeCycleException
  {
    super.unbindFc(clientItfName);
    if (clientItfName.equals(SEND_ACK_OPT_ITF_NAME))
    {
      sendAckOptItf = null;
    }
  }
}