/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.channel;

import java.io.IOException;
import java.util.Map;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.IOPushException;
import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.codec.MessageCodec;
import org.objectweb.dream.message.manager.MessageManager;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Generic Push ChannelOut. The {@link #push(Message, Map) push }method
 * retreive a socket using the bound {@link SocketManager }interface. it send
 * the message using the bound {@link MessageCodec }interface. Then if the
 * <code>"wait-ack-opt"</code> client interface is bound, it use it to wait an
 * acknowledgment from the ChannelIn. Finaly the socket is released and the
 * message is deleted.
 */
public class GenericPushChannelOutImpl extends AbstractComponent
    implements
      Push
{

  /**
   * The name of the optional client interface used by this coponent to wait for
   * an acknowlegdment
   */
  public static final String WAIT_ACK_OPT_ITF_NAME = "wait-ack-opt";

  // ---------------------------------------------------------------------------
  // Client interfaces
  // ---------------------------------------------------------------------------

  protected MessageManager   messageManagerItf;
  protected MessageCodec     messageCodecItf;
  protected SocketManager    socketManagerItf;
  protected WaitByte         waitAckOptItf;

  // ---------------------------------------------------------------------------
  // Implementation of the Push interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.Push#push(Message, Map)
   */
  public void push(Message message, Map context) throws PushException
  {
    if (logger.isLoggable(BasicLevel.DEBUG))
    {
      logger.log(BasicLevel.DEBUG, "try to send message -> " + message);
    }
    SocketState socketState = null;
    try
    {
      socketState = socketManagerItf.getSocket(message);
    }
    catch (InterruptedException e1)
    {
      throw new PushException(e1);
    }
    catch (IOException e)
    {
      throw new IOPushException(e);
    }
    try
    {
      logger.log(BasicLevel.DEBUG, "write message");
      messageCodecItf.encode(socketState, message);

      if (waitAckOptItf != null)
      {
        int b = waitAckOptItf.waitByte(socketState.getInput());
        if (b != 0)
        {
          socketManagerItf.releaseSocket(socketState, false);
          throw new PushException("Receive nack for message.");
        }
      }
    }
    catch (IOException e)
    {
      socketManagerItf.releaseSocket(socketState, true);
      throw new IOPushException(e);
    }

    socketManagerItf.releaseSocket(socketState, false);
    messageManagerItf.deleteMessage(message);
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{MessageManager.ITF_NAME, MessageCodec.ITF_NAME,
        SocketManager.ITF_NAME, WAIT_ACK_OPT_ITF_NAME};
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public synchronized void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }
    else if (clientItfName.equals(MessageCodec.ITF_NAME))
    {
      messageCodecItf = (MessageCodec) serverItf;
    }
    else if (clientItfName.equals(SocketManager.ITF_NAME))
    {
      socketManagerItf = (SocketManager) serverItf;
    }
    else if (clientItfName.equals(WAIT_ACK_OPT_ITF_NAME))
    {
      waitAckOptItf = (WaitByte) serverItf;
    }
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#unbindFc(String)
   */
  public void unbindFc(String clientItfName) throws NoSuchInterfaceException,
      IllegalBindingException, IllegalLifeCycleException
  {
    super.unbindFc(clientItfName);
    if (clientItfName.equals(WAIT_ACK_OPT_ITF_NAME))
    {
      waitAckOptItf = null;
    }
  }
}