/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.channel;

import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;
import java.net.SocketException;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.pool.ObjectPool;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * {@link IPSocketManager }that handles one TCP connection at a time. A new
 * connection is established for each
 * {@link IPSocketManager#getSocket(InetAddress, int) }method call, and
 * connections are closed by
 * {@link IPSocketManager#releaseSocket(SocketState, boolean) }method.
 */
public class IPSocketManagerTCPImpl extends AbstractComponent
    implements
      IPSocketManager,
      IPSocketManagerTCPAttributeController
{

  // ---------------------------------------------------------------------------
  // attribute fields
  // ---------------------------------------------------------------------------

  protected int              cnxRetry                   = CNX_RETRY_DEFAULT;
  protected boolean          tcpNoDelay                 = true;
  protected int              soTimeout                  = 0;
  protected int              soLinger                   = 60;

  // ---------------------------------------------------------------------------
  // Client interfaces
  // ---------------------------------------------------------------------------

  /** The name of the client interface used to pool socket state instances */
  public static final String SOCKET_STATE_POOL_ITF_NAME = "socket-state-pool";
  protected ObjectPool       socketStatePoolItf         = null;

  // ---------------------------------------------------------------------------
  // Implementation of IPSocketManager interface
  // ---------------------------------------------------------------------------

  /**
   * @see IPSocketManager#getSocket(InetAddress, int)
   */
  public SocketState getSocket(InetAddress address, int port)
      throws IOException
  {
    for (int i = 0;; i++)
    {
      try
      {
        logger.log(BasicLevel.DEBUG, "try to connect");
        Socket sock = new Socket(address, port);
        setSocketOption(sock);
        logger.log(BasicLevel.DEBUG, "connected");
        SocketState socketState = (SocketState) socketStatePoolItf
            .newInstance();
        socketState.setSocket(sock);
        return socketState;
      }
      catch (IOException exc)
      {
        if (i > cnxRetry)
        {
          throw exc;
        }
        logger.log(BasicLevel.WARN, "connection failed on address " + address
            + " will retry");
        try
        {
          Thread.sleep(i * 250);
        }
        catch (InterruptedException ignored)
        {
          // ignored
        }
      }
    }
  }

  /**
   * @see IPSocketManager#releaseSocket(SocketState, boolean)
   */
  public void releaseSocket(SocketState socketState, boolean error)
  {
    socketState.close();
    socketStatePoolItf.recycleInstance(socketState);
  }

  protected void setSocketOption(Socket sock) throws SocketException
  {
    sock.setTcpNoDelay(tcpNoDelay);
    sock.setSoTimeout(soTimeout);
    if (soLinger >= 0)
    {
      sock.setSoLinger(true, soLinger);
    }
    else
    {
      sock.setSoLinger(false, 0);
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of AttributeController interface
  // ---------------------------------------------------------------------------

  /**
   * @see IPSocketManagerTCPAttributeController#getCnxRetry()
   */
  public int getCnxRetry()
  {
    return cnxRetry;
  }

  /**
   * @see IPSocketManagerTCPAttributeController#setCnxRetry(int)
   */
  public void setCnxRetry(int cnxRetry)
  {
    this.cnxRetry = cnxRetry;
  }

  /**
   * @see IPSocketManagerTCPAttributeController#getTcpNoDelay()
   */
  public boolean getTcpNoDelay()
  {
    return tcpNoDelay;
  }

  /**
   * @see org.objectweb.dream.channel.IPSocketManagerTCPAttributeController#setTcpNoDelay(boolean)
   */
  public void setTcpNoDelay(boolean tcpNoDelay)
  {
    this.tcpNoDelay = tcpNoDelay;
  }

  /**
   * @see IPSocketManagerTCPAttributeController#getSoTimeout()
   */
  public int getSoTimeout()
  {
    return soTimeout;
  }

  /**
   * @see IPSocketManagerTCPAttributeController#setSoTimeout(int)
   */
  public void setSoTimeout(int timeout)
  {
    soTimeout = timeout;
  }

  /**
   * @see IPSocketManagerTCPAttributeController#getSoLinger()
   */
  public int getSoLinger()
  {
    return soLinger;
  }

  /**
   * @see IPSocketManagerTCPAttributeController#setSoLinger(int)
   */
  public void setSoLinger(int timeout)
  {
    soLinger = timeout;
  }

  // ---------------------------------------------------------------------------
  // Implementation of BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{SOCKET_STATE_POOL_ITF_NAME};
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(SOCKET_STATE_POOL_ITF_NAME))
    {
      socketStatePoolItf = (ObjectPool) serverItf;
    }
  }
}