/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): Matthieu Leclercq
 */

package org.objectweb.dream.control.activity.task;

import java.util.HashMap;
import java.util.Map;

import org.objectweb.dream.control.activity.scheduler.ForwarderSchedulerImpl;
import org.objectweb.dream.control.activity.scheduler.Scheduler;
import org.objectweb.dream.control.activity.task.thread.ThreadPoolController;
import org.objectweb.dream.control.activity.task.thread.ThreadTask;
import org.objectweb.dream.util.Error;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalContentException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.julia.InitializationContext;
import org.objectweb.fractal.util.Fractal;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Provides a basic implementation of the
 * {@link org.objectweb.dream.control.activity.task.TaskManagerController}
 * interface. Each registered task has its own scheduler and thread task. Two
 * types of thread tasks are available :
 * <ul>
 * <li>Mono threaded task (default): the {@link #registerTask}method returns
 * <code>null</code></li>
 * <li>Thread pool task: used if registration hints contain a
 * <code>"thread"</code> key mapped to the <code>"pool"</code> value. In
 * addition the pool capacity can be specified with a
 * <code>"threadPool.capacity"</code> key mapped to the required
 * {@link Integer }value. In this case, the {@link #registerTask}method
 * returns the
 * {@link org.objectweb.dream.control.activity.task.thread.ThreadPoolController}
 * control interface of the thread pool component.</li>
 * </ul>
 * <p>
 * 
 * @see org.objectweb.dream.control.activity.task.thread.ThreadTask
 * @see org.objectweb.dream.control.activity.task.thread.ThreadPoolTask
 * @see ThreadPoolController#setCapacity(int)
 */

public abstract class ThreadPerTaskTaskManagerControllerMixin
    implements
      TaskManagerController,
      TaskStoppedListener
{

  /** The map of registered tasks and the thread they are bound to. */
  protected Map tasks;
  /** The map of the interrupted tasks. */
  protected Map interruptedTasks;

  // ---------------------------------------------------------------------------
  // private constructor
  // ---------------------------------------------------------------------------

  private ThreadPerTaskTaskManagerControllerMixin()
  {
  }

  // ---------------------------------------------------------------------------
  // Fields and methods added and overriden by the mixin class
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.julia.Controller#initFcController(InitializationContext)
   */
  public void initFcController(final InitializationContext ic)
      throws InstantiationException
  {

    tasks = new HashMap();
    interruptedTasks = new HashMap();

    _super_initFcController(ic);
  }

  // ---------------------------------------------------------------------------
  // Implementation of the TaskManagerController interface
  // ---------------------------------------------------------------------------

  /**
   * @see TaskManagerController#registerTask(Task, Map)
   */
  public synchronized Object registerTask(Task task, Map hints)
      throws IllegalTaskException
  {
    if (tasks.containsKey(task))
    {
      throw new IllegalTaskException(task, "The task is already registred",
          null);
    }

    // Add task component.
    String controllerDesc = (String) hints.get("taskControllerDesc");
    Component componentTask;
    try
    {
      componentTask = Util.createTask(task, (controllerDesc == null)
          ? "taskPrimitive"
          : controllerDesc);
    }
    catch (InstantiationException e)
    {
      throw new IllegalTaskException(null,
          "Unable to instantiate task component", e);
    }
    try
    {
      _this_weaveableCC.addFcSubComponent(componentTask);
    }
    catch (IllegalContentException e)
    {
      throw new IllegalTaskException(task,
          "Unable to add task component in activity manager", e);
    }
    catch (IllegalLifeCycleException e)
    {
      throw new IllegalTaskException(task,
          "Unable to add task component in activity manager", e);
    }
    String taskName = null;
    try
    {
      taskName = Fractal.getNameController(componentTask).getFcName();
    }
    catch (NoSuchInterfaceException e1)
    {
      // ignore.
    }

    // Create the thread task component.
    Component thread;
    ThreadPoolController returnObject = null;
    String s = (String) hints.get("thread");
    if (s != null && "pool".equals(s))
    {
      _this_weaveableTMCLogger.log(BasicLevel.DEBUG, "Create thread pool task");

      try
      {
        thread = Util.createThreadPoolTask((taskName == null)
            ? "threadPoolTask"
            : taskName + "-threadPoolTask");
      }
      catch (InstantiationException e)
      {
        throw new IllegalTaskException(task,
            "Unable to create thread pool task component in activity manager",
            e);
      }

      try
      {
        returnObject = (ThreadPoolController) thread.getFcInterface("thread-pool-controller");
      }
      catch (NoSuchInterfaceException e)
      {
        Error.bug(_this_weaveableTMCLogger, e);
      }
      // set initial capacity;
      Integer capacity = (Integer) hints.get("threadPool.capacity");
      if (capacity != null)
      {
        returnObject.setCapacity(capacity.intValue());
      }
    }
    else
    {
      _this_weaveableTMCLogger
          .log(BasicLevel.DEBUG, "Create basic thread task");
      try
      {
        thread = Util.createTask(new ThreadTask(), "threadPrimitive");
      }
      catch (InstantiationException e)
      {
        throw new IllegalTaskException(task,
            "Unable to create thread task component in activity manager", e);
      }
      returnObject = null;
    }

    if (taskName != null)
    {
      try
      {
        Fractal.getNameController(thread).setFcName(taskName + "-thread");
      }
      catch (NoSuchInterfaceException e2)
      {
        // ignore.
      }
    }

    // Add the thread task component.
    try
    {
      _this_weaveableCC.addFcSubComponent(thread);
    }
    catch (IllegalContentException e)
    {
      throw new IllegalTaskException(task,
          "Unable to add thread task component in activity manager", e);
    }
    catch (IllegalLifeCycleException e)
    {
      throw new IllegalTaskException(task,
          "Unable to add thread task component in activity manager", e);
    }

    // Create and add the forwarder scheduler
    Component forwarderScheduler;
    try
    {
      forwarderScheduler = Util.createScheduler(new ForwarderSchedulerImpl(),
          (taskName == null) ? "scheduler" : taskName + "-scheduler");
    }
    catch (InstantiationException e)
    {
      throw new IllegalTaskException(task,
          "Unable to create scheduler component in activity manager", e);
    }
    if (taskName != null)
    {
      try
      {
        Fractal.getNameController(forwarderScheduler).setFcName(
            taskName + "-scheduler");
      }
      catch (NoSuchInterfaceException e2)
      {
        // ignore.
      }
    }

    try
    {
      _this_weaveableCC.addFcSubComponent(forwarderScheduler);
    }
    catch (IllegalContentException e)
    {
      throw new IllegalTaskException(task,
          "Unable to add scheduler component in activity manager", e);
    }
    catch (IllegalLifeCycleException e)
    {
      throw new IllegalTaskException(task,
          "Unable to add scheduler component in activity manager", e);
    }

    // bind the thread task to the scheduler
    try
    {
      Fractal.getBindingController(thread).bindFc(Scheduler.ITF_NAME,
          forwarderScheduler.getFcInterface(Scheduler.ITF_NAME));
      // bind the scheduler to the task
      Fractal.getBindingController(forwarderScheduler).bindFc(Task.ITF_NAME,
          componentTask.getFcInterface(Task.ITF_NAME));
    }
    catch (NoSuchInterfaceException e)
    {
      Error.bug(_this_weaveableTMCLogger, e);
    }
    catch (IllegalBindingException e)
    {
      Error.bug(_this_weaveableTMCLogger, e);
    }
    catch (IllegalLifeCycleException e)
    {
      Error.bug(_this_weaveableTMCLogger, e);
    }
    try
    {
      // start all the components
      Fractal.getLifeCycleController(thread).startFc();
    }
    catch (IllegalLifeCycleException e)
    {
      throw new IllegalTaskException(task,
          "unable to start thread task component", e);
    }
    catch (NoSuchInterfaceException e)
    {
      Error.bug(_this_weaveableTMCLogger, e);
    }

    tasks.put(task, new Component[]{componentTask, forwarderScheduler, thread});

    return returnObject;

  }

  /**
   * @see TaskManagerController#unregisterTask(Task)
   */
  public synchronized void unregisterTask(Task task)
      throws IllegalTaskException
  {
    if (!tasks.containsKey(task))
    {
      throw new IllegalTaskException(task, "The task is not registered", null);
    }
    // retrieve the forwarder scheduler and the thread associated to the
    // task
    Component[] components = (Component[]) tasks.get(task);
    Component componentTask = components[0];
    Component forwarderScheduler = components[1];
    Component thread = components[2];
    removeTask(task, componentTask, forwarderScheduler, thread);
    tasks.remove(task);
  }

  /**
   * @see TaskManagerController#interruptTask(Task, TaskStoppedListener)
   */
  public synchronized void interruptTask(Task task, TaskStoppedListener listener)
      throws IllegalTaskException
  {
    if (!tasks.containsKey(task))
    {
      throw new IllegalTaskException(task, "The task is not registered", null);
    }
    // retrieve the forwarder scheduler and the thread associated to the
    // task
    Component[] components = (Component[]) tasks.get(task);
    Component thread = components[2];
    Object[] taskStatus = new Object[]{task, listener, components[0],
        components[1]};
    interruptedTasks.put(thread, taskStatus);
    try
    {
      ((TaskLifeCycleController) Fractal.getLifeCycleController(thread))
          .asyncStop(this);
    }
    catch (NoSuchInterfaceException e)
    {
      // can't happend since thread has already been started.
      Error.bug(_this_weaveableTMCLogger, e);
    }
  }

  /**
   * @see TaskManagerController#getTasks()
   */
  public Task[] getTasks()
  {
    if (tasks == null)
    {
      return new Task[0];
    }
    return (Task[]) tasks.keySet().toArray(new Task[tasks.size()]);
  }

  // ---------------------------------------------------------------------------
  // Implementation of the TaskManagerController interface
  // ---------------------------------------------------------------------------

  /**
   * @see TaskStoppedListener#taskStopped(Task)
   */
  public void taskStopped(Task thread)
  {
    Task task;
    TaskStoppedListener listener;
    synchronized (this)
    {
      Component threadComponent = ((Interface) thread).getFcItfOwner();
      Object[] taskStatus = (Object[]) interruptedTasks.get(threadComponent);
      task = (Task) taskStatus[0];
      listener = (TaskStoppedListener) taskStatus[1];
      Component componentTask = (Component) taskStatus[2];
      Component forwarderScheduler = (Component) taskStatus[3];
      try
      {
        removeTask(task, componentTask, forwarderScheduler, threadComponent);
      }
      catch (IllegalTaskException e)
      {
        Error.bug(_this_weaveableTMCLogger, e);
      }
      interruptedTasks.remove(threadComponent);
    }
    listener.taskStopped(task);
  }

  // ---------------------------------------------------------------------------
  // Utility method
  // ---------------------------------------------------------------------------

  /**
   * Removes a task from this composite component.
   * 
   * @param task the registered task.
   * @param componentTask the task component.
   * @param forwarderScheduler the scheduler of the task.
   * @param thread the thread executing the task.
   * @throws IllegalTaskException if an error occurs.
   */
  public void removeTask(Task task, Component componentTask,
      Component forwarderScheduler, Component thread)
      throws IllegalTaskException
  {
    try
    {
      // stop, unbind and remove the thread
      Fractal.getLifeCycleController(thread).stopFc();
    }
    catch (IllegalLifeCycleException e)
    {
      throw new IllegalTaskException(task,
          "unable to stop thread task component", e);
    }
    catch (NoSuchInterfaceException e)
    {
      Error.bug(_this_weaveableTMCLogger, e);
    }
    // stop, unbind and remove the scheduler
    try
    {
      Fractal.getBindingController(thread).unbindFc(Scheduler.ITF_NAME);
      _this_weaveableCC.removeFcSubComponent(thread);
      Fractal.getBindingController(forwarderScheduler).unbindFc(Task.ITF_NAME);
      _this_weaveableCC.removeFcSubComponent(forwarderScheduler);
      // remove the task
      _this_weaveableCC.removeFcSubComponent(componentTask);
    }
    catch (NoSuchInterfaceException e)
    {
      Error.bug(_this_weaveableTMCLogger, e);
    }
    catch (IllegalBindingException e)
    {
      Error.bug(_this_weaveableTMCLogger, e);
    }
    catch (IllegalLifeCycleException e)
    {
      Error.bug(_this_weaveableTMCLogger, e);
    }
    catch (IllegalContentException e)
    {
      Error.bug(_this_weaveableTMCLogger, e);
    }
  }

  // ---------------------------------------------------------------------------
  // Fields and methods required by the mixin class in the base class
  // ---------------------------------------------------------------------------

  /**
   * The <tt>weaveableTMCLogger</tt> field required by this mixin. This field
   * is supposed to reference the {@link Logger }of this controller.
   */
  public Logger            _this_weaveableTMCLogger;

  /**
   * The {@link ContentController }interface of the component to which this
   * controller object belongs.
   */
  public ContentController _this_weaveableCC;

  /**
   * The {@link org.objectweb.fractal.julia.Controller#initFcController }method
   * overriden by this mixin.
   * 
   * @see org.objectweb.fractal.julia.Controller#initFcController(InitializationContext)
   */
  public abstract void _super_initFcController(InitializationContext ic)
      throws InstantiationException;
}