/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.message;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * Extensible message implementation specialized for aggregate message (ie.
 * message containing only sub messages).
 */
public class AggregateMessageImpl
    implements
      ExtensibleMessage,
      MessageReferenceCounter
{
  final List      subMessages = new ArrayList();
  transient short messageManagerId;
  transient int   referenceCounter;

  // ---------------------------------------------------------------------------
  //  Implementation of the ExtensibleMessage interface
  // ---------------------------------------------------------------------------

  /**
   * Throws a {@link UnsupportedOperationException}.
   * 
   * @see ExtensibleMessage#addChunk(String, ChunkType, Object)
   */
  public void addChunk(String name, ChunkType chkType, Object chunk)
      throws ChunkAlreadyExistsException
  {
    throw new UnsupportedOperationException();
  }

  /**
   * Always return <code>null</code>.
   * 
   * @see ExtensibleMessage#removeChunk(String)
   */
  public Object removeChunk(String name)
  {
    return null;
  }

  /**
   * @see ExtensibleMessage#addSubMessage(Message)
   */
  public void addSubMessage(Message message)
  {
    subMessages.add(message);
  }

  /**
   * @see ExtensibleMessage#removeSubMessage(Message)
   */
  public boolean removeSubMessage(Message message)
  {
    return subMessages.remove(message);
  }

  /**
   * @see ExtensibleMessage#removeSubMessages()
   */
  public void removeSubMessages()
  {
    subMessages.clear();
  }

  // ---------------------------------------------------------------------------
  //  Implementation of the Message interface
  // ---------------------------------------------------------------------------

  /**
   * @see Message#getMessageManagerId()
   */
  public short getMessageManagerId()
  {
    return messageManagerId;
  }

  /**
   * @see Message#setMessageManagerId(short)
   */
  public void setMessageManagerId(short id)
  {
    messageManagerId = id;
  }

  /**
   * @see Message#getChunk(String)
   */
  public Object getChunk(String name)
  {
    return null;
  }

  /**
   * @see Message#getSubMessageIterator()
   */
  public Iterator getSubMessageIterator()
  {
    return subMessages.iterator();
  }

  /**
   * @see Message#getSubMessages()
   */
  public Message[] getSubMessages()
  {
    Message[] subMessageArray = new Message[subMessages.size()];
    return (Message[]) subMessages.toArray(subMessageArray);
  }

  /**
   * Throws a {@link UnsupportedOperationException}.
   * 
   * @see Message#getMessageType()
   */
  public MessageType getMessageType()
  {
    throw new UnsupportedOperationException();
  }

  /**
   * @see Message#transfertChunkStates(Message)
   */
  public void transfertChunkStates(Message newInstance)
  {
    // nothing to do since this message contains no chunk
  }

  /**
   * @see Message#recycle()
   */
  public void recycle()
  {
    subMessages.clear();
  }

  // ---------------------------------------------------------------------------
  //  Implementation of the MessageReferenceCounter interface
  // ---------------------------------------------------------------------------

  /**
   * @see MessageReferenceCounter#incrementReferenceCounter()
   */
  public void incrementReferenceCounter()
  {
    referenceCounter++;
  }

  /**
   * @see MessageReferenceCounter#decrementReferenceCounter()
   */
  public boolean decrementReferenceCounter()
  {
    synchronized (this)
    {
      referenceCounter--;
      return (referenceCounter == 0);
    }
  }

}