/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): Vivien Quema
 */

package org.objectweb.dream.message.manager;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.message.AbstractExtensibleMessage;
import org.objectweb.dream.message.Chunk;
import org.objectweb.dream.message.ChunkAlreadyExistsException;
import org.objectweb.dream.message.ChunkType;
import org.objectweb.dream.message.ChunkTypeImpl;
import org.objectweb.dream.message.ExtensibleMessage;
import org.objectweb.dream.message.ExtensibleMessageImpl;
import org.objectweb.dream.message.ExtensibleMessageNC;
import org.objectweb.dream.message.ExtensibleMessageNCImpl;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageAlreadyExistException;
import org.objectweb.dream.message.MessageReferenceCounter;
import org.objectweb.dream.message.MessageType;
import org.objectweb.dream.message.MessageTypeNC;
import org.objectweb.dream.pool.ObjectPool;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Basic implementation of a message manager.
 */
public class MessageManagerImpl extends AbstractComponent
    implements
      MessageManager,
      MessageManagerAttributeController
{

  // ---------------------------------------------------------------------------
  // Attribute field
  // ---------------------------------------------------------------------------

  // the id of this message manager
  short                      id                                      = -1;

  // ---------------------------------------------------------------------------
  // Client interfaces
  // ---------------------------------------------------------------------------

  /**
   * The name of the optional client interface used to pool
   * {@link ExtensibleMessage }instances.
   */
  public static final String EXTENSIBLE_MESSAGE_POOL_OPT_ITF_NAME    = "extensible-message-pool";

  /**
   * The name of the optional client interface used to pool
   * {@link ExtensibleMessageNC }instances.
   */
  public static final String EXTENSIBLE_MESSAGE_NC_POOL_OPT_ITF_NAME = "extensible-message-NC-pool";

  /**
   * The name of the collection interface used to pool chunk instances
   */
  public static final String CHUNK_POOL_COLLECTION_ITF_NAME          = "chunk-pool";

  protected ObjectPool       extensibleMessagePoolOptItf             = null;
  protected ObjectPool       extensibleMessageNCPoolOptItf           = null;

  protected Map              chunkPoolItfs                           = null;
  protected List             unknownChunkPools                       = null;

  // ---------------------------------------------------------------------------
  // Implementation of the MessageManager interface
  // ---------------------------------------------------------------------------

  /**
   * @see MessageManager#createMessage(MessageType)
   */
  public Message createMessage(MessageType type) throws UnknownChunkTypeError
  {
    ExtensibleMessage msg;
    Iterator iterator;
    if (type instanceof MessageTypeNC)
    {
      ExtensibleMessageNC message = createExtensibleMessageNC();
      msg = message;
      MessageTypeNC typeNC = (MessageTypeNC) type;

      // add named sub messages
      iterator = typeNC.getSubMessageNamesIterator();
      while (iterator.hasNext())
      {
        String msgName = (String) iterator.next();
        MessageType t = typeNC.getSubMessageType(msgName);
        try
        {
          message.addSubMessage(msgName, createMessage(t));
        }
        catch (MessageAlreadyExistException e)
        {
          // cannot happend
        }
      }

      iterator = typeNC.getUnnamedSubMessageTypesIterator();
    }
    else
    {
      msg = createExtensibleMessage();
      iterator = type.getSubMessageTypesIterator();
    }
    // add unnamed sub messages
    while (iterator.hasNext())
    {
      MessageType t = (MessageType) iterator.next();
      msg.addSubMessage(createMessage(t));
    }

    // add chunks
    iterator = type.getChunkNamesIterator();
    while (iterator.hasNext())
    {
      String chunkName = (String) iterator.next();
      ChunkType t = type.getChunkType(chunkName);
      try
      {
        msg.addChunk(chunkName, t, createChunk(t));
      }
      catch (ChunkAlreadyExistsException e)
      {
        // cannot happend
      }
    }
    return msg;
  }

  /**
   * @see MessageManager#deleteMessage(Message)
   */
  public void deleteMessage(Message message)
  {
    if (((MessageReferenceCounter) message).decrementReferenceCounter())
    {
      logger.log(BasicLevel.DEBUG, "Recycle message.");
      Iterator iterator;
      // if no chunk pool, it is not necessary to recycle chunk instance
      if (chunkPoolItfs != null)
      {
        iterator = message.getMessageType().getChunkNamesIterator();
        while (iterator.hasNext())
        {
          String chunkName = (String) iterator.next();
          deleteChunk(message.getChunk(chunkName));
        }
      }

      iterator = message.getSubMessageIterator();
      while (iterator.hasNext())
      {
        Message subMessage = (Message) iterator.next();
        deleteMessage(subMessage);
      }
      if (message instanceof ExtensibleMessageNC)
      {
        if (extensibleMessageNCPoolOptItf != null)
        {
          extensibleMessageNCPoolOptItf.recycleInstance(message);
        }
      }
      else
      {
        if (extensibleMessagePoolOptItf != null)
        {
          extensibleMessagePoolOptItf.recycleInstance(message);
        }
      }
    }
  }

  /**
   * @see MessageManager#duplicateMessage(Message, boolean)
   */
  public Message duplicateMessage(Message message, boolean clone)
  {
    if (clone)
    {
      // TODO wrong for sub message !!
      Message newInstance = createMessage(message.getMessageType());
      message.transfertChunkStates(newInstance);
      return newInstance;
    }
    else
    {
      ((AbstractExtensibleMessage) message).incrementReferenceCounter();
      return message;
    }
  }

  /**
   * @see MessageManager#createChunk(ChunkType)
   */
  public Object createChunk(ChunkType type) throws UnknownChunkTypeError
  {
    if (type instanceof ChunkTypeImpl)
    {
      Chunk chunk = createChunkInstance(type);
      chunk.setMessageManagerId(id);
      return chunk;
    }
    else
    {
      throw new UnknownChunkTypeError(type);
    }
  }

  /**
   * @see MessageManager#deleteChunk(Object)
   */
  public void deleteChunk(Object chunk)
  {
    if (chunk instanceof Chunk)
    {
      recycleChunkInstance((Chunk) chunk);
    }
  }

  /**
   * @see MessageManager#duplicateChunk(Object, boolean)
   */
  public Object duplicateChunk(Object chunk, boolean clone)
  {
    if (chunk instanceof Chunk)
    {
      Chunk c = (Chunk) chunk;
      Chunk newInstance = (Chunk) createChunk(c.getType());
      c.transfertState(newInstance);
      return c;
    }
    else
    {
      throw new UnsupportedOperationException(
          "Unable to duplicate chunk, does not implement Chunk interface : "
              + chunk);
    }
  }

  /**
   * @see MessageManager#getMessageManagerId()
   */
  public short getMessageManagerId()
  {
    return id;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the MessageManagerAttributeController
  // ---------------------------------------------------------------------------

  /**
   * @see MessageManagerAttributeController#getId()
   */
  public short getId()
  {
    return id;
  }

  /**
   * @see MessageManagerAttributeController#setId(short)
   */
  public void setId(short id)
  {
    this.id = id;
  }

  // ---------------------------------------------------------------------------
  // Utilty methods
  // ---------------------------------------------------------------------------

  protected final ExtensibleMessageNC createExtensibleMessageNC()
  {
    if (extensibleMessageNCPoolOptItf == null)
    {
      return new ExtensibleMessageNCImpl(id);
    }
    else
    {
      return (ExtensibleMessageNC) extensibleMessageNCPoolOptItf.newInstance();
    }
  }

  protected final ExtensibleMessage createExtensibleMessage()
  {
    if (extensibleMessagePoolOptItf == null)
    {
      return new ExtensibleMessageImpl(id);
    }
    else
    {
      return (ExtensibleMessage) extensibleMessagePoolOptItf.newInstance();
    }
  }

  protected final Chunk createChunkInstance(ChunkType chunkType)
  {
    if (unknownChunkPools != null)
    {
      if (chunkPoolItfs == null)
      {
        chunkPoolItfs = new HashMap();
      }
      // their is unknown chunk pool binding.
      Iterator iterator = unknownChunkPools.iterator();
      while (iterator.hasNext())
      {
        ObjectPool chunkPool = (ObjectPool) iterator.next();
        Chunk instance = (Chunk) chunkPool.newInstance();
        ChunkType type = instance.getType();
        chunkPoolItfs.put(type, instance);
        chunkPool.recycleInstance(instance);
      }
      unknownChunkPools = null; // gc the List
    }
    if (chunkPoolItfs != null)
    {
      ObjectPool pool = (ObjectPool) chunkPoolItfs.get(chunkType);
      if (pool != null)
      {
        return (Chunk) pool.newInstance();
      }
    }
    try
    {
      return (Chunk) ((ChunkTypeImpl) chunkType).getChunkImpl().newInstance();
    }
    catch (Exception e)
    {
      throw new UnknownChunkTypeError(
          "Unable to instanciate chunk implementation : " + e.getMessage(),
          chunkType);
    }
  }

  protected final void recycleChunkInstance(Chunk chunk)
  {
    if (chunkPoolItfs != null)
    {
      ObjectPool pool = (ObjectPool) chunkPoolItfs.get(chunk.getType());
      if (pool != null)
      {
        pool.recycleInstance(chunk);
      }
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(EXTENSIBLE_MESSAGE_POOL_OPT_ITF_NAME))
    {
      extensibleMessagePoolOptItf = (ObjectPool) serverItf;
    }
    else if (clientItfName.equals(EXTENSIBLE_MESSAGE_NC_POOL_OPT_ITF_NAME))
    {
      extensibleMessageNCPoolOptItf = (ObjectPool) serverItf;
    }
    else if (clientItfName.startsWith(CHUNK_POOL_COLLECTION_ITF_NAME))
    {
      if (chunkPoolItfs == null)
      {
        chunkPoolItfs = new HashMap();
      }
      chunkPoolItfs.put(clientItfName, serverItf);
      if (unknownChunkPools == null)
      {
        unknownChunkPools = new ArrayList();
      }
      unknownChunkPools.add(serverItf);
    }
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#unbindFc(String)
   */
  public void unbindFc(String clientItfName) throws NoSuchInterfaceException,
      IllegalBindingException, IllegalLifeCycleException
  {
    if (clientItfName.equals(EXTENSIBLE_MESSAGE_POOL_OPT_ITF_NAME))
    {
      extensibleMessagePoolOptItf = null;
    }
    else if (clientItfName.equals(EXTENSIBLE_MESSAGE_NC_POOL_OPT_ITF_NAME))
    {
      extensibleMessageNCPoolOptItf = null;
    }
    else if (clientItfName.startsWith(CHUNK_POOL_COLLECTION_ITF_NAME))
    {
      Object serverItf = super.lookupFc(clientItfName);
      if (unknownChunkPools != null)
      {
        unknownChunkPools.remove(serverItf);
      }
      chunkPoolItfs.values().remove(serverItf);
    }
    super.unbindFc(clientItfName);
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{EXTENSIBLE_MESSAGE_NC_POOL_OPT_ITF_NAME,
        EXTENSIBLE_MESSAGE_POOL_OPT_ITF_NAME, CHUNK_POOL_COLLECTION_ITF_NAME};
  }
}