/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.utobcast;

import java.util.Map;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.IOPushException;
import org.objectweb.dream.InterruptedPushException;
import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.PushWithReturn;
import org.objectweb.dream.message.ChunkAlreadyExistsException;
import org.objectweb.dream.message.ExtensibleMessage;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.manager.MessageManager;
import org.objectweb.dream.protocol.Process;
import org.objectweb.dream.protocol.utobcast.message.CrashedLeaderChunk;
import org.objectweb.dream.protocol.utobcast.message.UTOBcastChunk;
import org.objectweb.dream.util.Error;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Implementation of the UTOBroadcast component. This component sends messages
 * to the leader. If it detects a failure of the leader, it sends the message to
 * the backup.
 */
public class BroadcastImpl extends AbstractComponent implements Push
{

  // ---------------------------------------------------------------------------
  // Fields of this class
  // ---------------------------------------------------------------------------

  /** The interface to send messages to be delivered. */
  protected Push              outPushItf;

  /** The interface to send messages to be UTO broadcast. */
  protected PushWithReturn    outPushWithReturnItf;

  /** The interface to retrieve process membership information. */
  protected ProcessMembership processMembershipItf;

  /** The interface to manage message lifecycle. */
  protected MessageManager    messageManagerItf;

  Process                     myProcess;
  boolean                     configured = false;

  // ---------------------------------------------------------------------------
  // Constructor
  // ---------------------------------------------------------------------------

  /**
   * Constructor.
   */
  public BroadcastImpl()
  {
  }

  // ---------------------------------------------------------------------------
  // Implementation of the Push interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.Push#push(org.objectweb.dream.message.Message,
   *      java.util.Map)
   */
  public void push(Message message, Map context) throws PushException
  {
    if (logger.isLoggable(BasicLevel.INFO))
    {
      logger.log(BasicLevel.INFO, "Received message " + message
          + " to be broadcast");
    }
    if (!configured)
    {
      try
      {
        configure();
      }
      catch (InterruptedException e1)
      {
        throw new InterruptedPushException(e1);
      }
    }
    // Retrieve the UTOBcast chunk
    UTOBcastChunk utobcastChunk = (UTOBcastChunk) message
        .getChunk(UTOBcastChunk.DEFAULT_NAME);
    if (utobcastChunk == null)
    {
      // UTOBcast chunk not present -> create it
      if (message instanceof ExtensibleMessage)
      {
        utobcastChunk = (UTOBcastChunk) messageManagerItf
            .createChunk(UTOBcastChunk.TYPE);
        try
        {
          ((ExtensibleMessage) message).addChunk(UTOBcastChunk.DEFAULT_NAME,
              UTOBcastChunk.TYPE, utobcastChunk);
        }
        catch (ChunkAlreadyExistsException e2)
        {
          // Cannot happen
          Error.bug(logger, e2);
        }
      }
      else
      {
        // The message is not extensible -> unable to add a chunk
        Error.error("Message is not extensible: unable to add UTOBcast chunk",
            logger);
      }
    }

    // Fill the UTOBcast chunk (message type and sender)
    utobcastChunk.setUTOBcastMessageType(UTOBcastChunk.DAT);
    utobcastChunk.setProcessFrom(myProcess);

    Message returnMessage = null;

    while (returnMessage == null)
    {
      //    Retrieve the leader process
      Process leader;
      try
      {
        leader = processMembershipItf.getLeader();
      }
      catch (InterruptedException e1)
      {
        throw new InterruptedPushException(e1);
      }
      utobcastChunk.setProcessTo(leader);
      try
      {
        //Send the message
        if (logger.isLoggable(BasicLevel.INFO))
        {
          logger.log(BasicLevel.INFO, "Send DAT message " + message
              + " to leader");
        }
        returnMessage = outPushWithReturnItf.pushWithReturn(message, null);
        if (returnMessage.getChunk(CrashedLeaderChunk.DEFAULT_NAME) != null)
        {
          // We are notified that the leader has crashed
          // new leader should already have been elected
          if (logger.isLoggable(BasicLevel.INFO))
          {
            logger.log(BasicLevel.INFO,
                "Received AllKey -> the leader has crahsed");
          }
          returnMessage = null;
          continue;
        }
        else
        {
          // Deliver the message to myself
          if (logger.isLoggable(BasicLevel.INFO))
          {
            logger.log(BasicLevel.INFO, "Received message to be delivered "
                + returnMessage);
          }
          outPushItf.push(returnMessage, null);
        }
      }
      catch (IOPushException e)
      {
        // We detect that the leader has crashed (returnMessage still null)
        processMembershipItf.electBackupAsLeader(leader);
      }
    }
  }

  // ---------------------------------------------------------------------------
  // Utility methods.
  // ---------------------------------------------------------------------------

  void configure() throws InterruptedException
  {
    myProcess = processMembershipItf.getMyself();
    configured = true;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(java.lang.String,
   *      java.lang.Object)
   */
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(Push.OUT_PUSH_ITF_NAME))
    {
      outPushItf = (Push) serverItf;
    }
    else if (clientItfName.equals(PushWithReturn.OUT_PUSH_WITH_RETURN_ITF_NAME))
    {
      outPushWithReturnItf = (PushWithReturn) serverItf;
    }
    else if (clientItfName.equals(ProcessMembership.ITF_NAME))
    {
      processMembershipItf = (ProcessMembership) serverItf;
    }
    else if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{Push.OUT_PUSH_ITF_NAME,
        PushWithReturn.OUT_PUSH_ITF_NAME, ProcessMembership.ITF_NAME,
        MessageManager.ITF_NAME};
  }

}