/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.utobcast;

import java.util.HashMap;
import java.util.TreeSet;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.protocol.Process;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Implementation of the {@link ProcessMembership}interface.
 */
public class ProcessMembershipImpl extends AbstractComponent
    implements
      ProcessMembership,
      ProcessMembershipAttributeController
{

  boolean   blocking                                = false;

  // ---------------------------------------------------------------------------
  // Fields of this class
  // ---------------------------------------------------------------------------

  /**
   * The collection client interface bound to components that need be notified
   * about updates.
   */
  HashMap   processMembershipUpdateNotificationsItf = new HashMap();

  /** The leader process. */
  Process   leader;
  Object    leaderLock                              = new Object();

  /** The backup process. */
  Process   backup;
  Object    backupLock                              = new Object();

  /** The process that contains this component. */
  Process   myself;
  Object    myselfLock                              = new Object();

  /**
   * A TreeMapBuffer containing all the processes except leader and backup
   * processes.
   */
  TreeSet   otherProcesses                          = new TreeSet();
  Object    otherProcessesLock                      = new Object();

  /** An array containing all the processes except leader and backup processes. */
  Process[] otherProcessesArray;

  // ---------------------------------------------------------------------------
  // Constructor
  // ---------------------------------------------------------------------------

  /**
   * Constructor.
   */
  public ProcessMembershipImpl()
  {
  }

  // ---------------------------------------------------------------------------
  // Implementation of the ProcessMembership interface
  // ---------------------------------------------------------------------------

  /**
   * @throws InterruptedException
   * @see org.objectweb.dream.protocol.utobcast.ProcessMembership#getLeader()
   */
  public Process getLeader() throws InterruptedException
  {
    if (leader == null)
    {
      if (blocking)
      {
        synchronized (leaderLock)
        {
          // To avoid taking the lock each time the method is called
          if (leader == null)
          {
            leaderLock.wait();
          }
        }
      }
    }
    return leader;
  }

  /**
   * @see ProcessMembership#setLeader(Process)
   */
  public void setLeader(Process leader)
  {
    synchronized (leaderLock)
    {
      this.leader = leader;
      leaderLock.notifyAll();
    }
  }

  /**
   * @see ProcessMembership#getBackup()
   */
  public Process getBackup() throws InterruptedException
  {
    if (backup == null)
    {
      if (blocking)
      {
        synchronized (backupLock)
        {
          // To avoid taking the lock each time the method is called
          if (backup == null)
          {
            backupLock.wait();
          }
        }
      }
    }
    return backup;
  }

  /**
   * @see ProcessMembership#setBackup(Process)
   */
  public void setBackup(Process backup)
  {
    synchronized (backupLock)
    {
      this.backup = backup;
      backupLock.notifyAll();
    }
  }

  /**
   * @throws InterruptedException
   * @see ProcessMembership#getMyself()
   */
  public Process getMyself() throws InterruptedException
  {
    if (myself == null)
    {
      if (blocking)
      {
        synchronized (myselfLock)
        {
          // To avoid taking the lock each time the method is called
          if (myself == null)
          {
            myselfLock.wait();
          }
        }
      }
    }
    return myself;
  }

  /**
   * @see ProcessMembership#setMyself(Process)
   */
  public void setMyself(Process myself)
  {
    synchronized (myselfLock)
    {
      this.myself = myself;
      myselfLock.notifyAll();
    }
  }

  /**
   * @see org.objectweb.dream.protocol.utobcast.ProcessMembership#setOtherProcesses(org.objectweb.dream.protocol.Process[])
   */
  public void setOtherProcesses(Process[] processes)
  {
    synchronized (otherProcessesLock)
    {
      otherProcesses = new TreeSet();
      for (int i = 0; i < processes.length; i++)
      {
        otherProcesses.add(processes[i]);
      }
      this.otherProcessesArray = processes;
      otherProcessesLock.notifyAll();
    }
  }

  /**
   * @see ProcessMembership#getOtherProcesses()
   */
  public Process[] getOtherProcesses()
  {
    synchronized (otherProcessesLock)
    {
      return otherProcessesArray;
    }
  }

  /**
   * @see ProcessMembership#addProcess(Process)
   */
  public void addProcess(Process process)
  {
    synchronized (otherProcessesLock)
    {
      otherProcesses.add(process);
      createOtherProcessesArray();
      otherProcessesLock.notifyAll();
    }
  }

  /**
   * @see ProcessMembership#removeProcess(org.objectweb.dream.protocol.Process)
   */
  public void removeProcess(Process process)
  {
    synchronized (otherProcessesLock)
    {
      otherProcesses.remove(process);
      createOtherProcessesArray();
    }
  }

  /**
   * @see org.objectweb.dream.protocol.utobcast.ProcessMembership#electBackup()
   */
  public Process electBackup() throws InterruptedException
  {
    synchronized (otherProcessesLock)
    {
      while (otherProcesses.size() <= 0)
      {
        if (blocking)
        {
          otherProcessesLock.wait();
        }
        else
        {
          return null;
        }
      }
      backup = (Process) otherProcesses.first();
      otherProcesses.remove(backup);
      createOtherProcessesArray();
      return backup;
    }
  }

  /**
   * @see org.objectweb.dream.protocol.utobcast.ProcessMembership#electBackupAsLeader(org.objectweb.dream.protocol.Process)
   */
  public Process electBackupAsLeader(Process oldLeader)
  {
    synchronized (leaderLock)
    {
      if (oldLeader.equals(leader))
      {
        logger.log(BasicLevel.INFO, "elect the backup as leader");
        leader = backup;
        // Not necessary (nobody should be waiting)
        leaderLock.notifyAll();
      }
    }
    return leader;
  }

  // ---------------------------------------------------------------------------
  // Private methods
  // ---------------------------------------------------------------------------

  /**
   * Creates the <code>otherProcessesArray</code> array. Calls to this method
   * must be synchronized with <code>this</code>.
   */
  private void createOtherProcessesArray()
  {
    otherProcessesArray = new Process[otherProcesses.size()];
    otherProcesses.toArray(otherProcessesArray);
  }

  // ---------------------------------------------------------------------------
  // Implementation of the ProcessMembershipAttributeController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.protocol.utobcast.ProcessMembershipAttributeController#setBlocking(boolean)
   */
  public void setBlocking(boolean blocking)
  {
    this.blocking = blocking;
  }

  /**
   * @see org.objectweb.dream.protocol.utobcast.ProcessMembershipAttributeController#getBlocking()
   */
  public boolean getBlocking()
  {
    return blocking;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(java.lang.String,
   *      java.lang.Object)
   */
  public synchronized void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    processMembershipUpdateNotificationsItf.put(clientItfName, serverItf);
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#unbindFc(java.lang.String)
   */
  public synchronized void unbindFc(String clientItfName)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.unbindFc(clientItfName);
    processMembershipUpdateNotificationsItf.remove(clientItfName);
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public synchronized String[] listFc()
  {
    String[] toBeReturned = new String[processMembershipUpdateNotificationsItf
        .size()];
    processMembershipUpdateNotificationsItf.keySet().toArray(toBeReturned);
    return toBeReturned;
  }

}