/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.utobcast;

import java.util.Map;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.IOPushException;
import org.objectweb.dream.InterruptedPullException;
import org.objectweb.dream.InterruptedPushException;
import org.objectweb.dream.Pull;
import org.objectweb.dream.PullException;
import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.manager.MessageManager;
import org.objectweb.dream.protocol.utobcast.message.UTOBcastChunk;
import org.objectweb.dream.util.Error;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Implementation of the UTO component. This component handles UTO messages.
 * Upon message reception,
 * <ul>
 * <li>either it delivers the message and all the messages stored in the
 * <code>toBeDelivered</code> queue that can be delivered.</li>
 * <li>or it stores the message in the <code>toBeDelivered</code>.</li>
 * </ul>
 * In both cases, an ACK is sent to pleader and pbackup.
 */
public class UTOImpl extends AbstractComponent implements Push
{

  // ---------------------------------------------------------------------------
  // Fields of this class
  // ---------------------------------------------------------------------------

  /**
   * The commonly used name to refer to the <code>deliveryItf</code>
   * interface.
   */
  public static final String  DELIVERY_ITF_NAME            = "delivery";

  /**
   * The commonly used name to refer to the <code>toBeDeliveredInItf</code>
   * interface.
   */
  public static final String  TO_BE_DELIVERED_IN_ITF_NAME  = "to-be-delivered-in";

  /**
   * The commonly used name to refer to the <code>toBeDeliveredOutItf</code>
   * interface.
   */
  public static final String  TO_BE_DELIVERED_OUT_ITF_NAME = "to-be-delivered-out";

  /** The interface to deliver messages. */
  protected Push              deliveryItf;

  /** The interface to store messages in the <code>ToBeDelivered</code> queue. */
  protected Push              toBeDeliveredInItf;

  /**
   * The interface to retrieve messages to be delivered from the
   * <code>toBeDelivered</code> queue.
   */
  protected Pull              toBeDeliveredOutItf;

  /** The interface to send ACK messages. */
  protected Push              outPushItf;

  /** The interface to retrieve process membership information. */
  protected ProcessMembership processMembershipItf;

  /** The interface to manage message lifecycle. */
  protected MessageManager    messageManagerItf;

  // ---------------------------------------------------------------------------
  // Constructor
  // ---------------------------------------------------------------------------

  /**
   * Constructor.
   */
  public UTOImpl()
  {
  }

  // ---------------------------------------------------------------------------
  // Implementation of the Push interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.Push#push(org.objectweb.dream.message.Message,
   *      java.util.Map)
   */
  public void push(Message message, Map context) throws PushException
  {
    if (logger.isLoggable(BasicLevel.INFO))
    {
      logger.log(BasicLevel.INFO, "Received UTO message " + message);
    }
    //////////////////////////////////////////////////////////
    // Add the message to the ToBeDelivered list
    //////////////////////////////////////////////////////////

    if (logger.isLoggable(BasicLevel.INFO))
    {
      logger.log(BasicLevel.INFO, "Add UTO message " + message
          + " to the toBeDelivered queue");
    }
    Message msg = messageManagerItf.duplicateMessage(message, true);
    toBeDeliveredInItf.push(msg, null);

    //////////////////////////////////////////////////////////
    // Deliver all messages that are stored in the ToBeDelivered list and that
    // can be delivered
    //////////////////////////////////////////////////////////
    try
    {
      while ((msg = toBeDeliveredOutItf.pull(null)) != null)
      {
        if (logger.isLoggable(BasicLevel.INFO))
        {
          logger.log(BasicLevel.INFO, "Deliver UTO message " + msg);
        }
        deliveryItf.push(msg, null);
      }
    }
    catch (InterruptedPullException e)
    {
      // Happens if the queue is in blocking mode
      throw new InterruptedPushException("Interrupted while pulling a message");
    }
    catch (PullException e)
    {
      // This is a bug: queue must either return null or block
      Error.bug(logger, e);
    }
    //////////////////////////////////////////////////////////
    // Send ACK message to leader and backup
    //////////////////////////////////////////////////////////

    UTOBcastChunk chunk = (UTOBcastChunk) message
        .getChunk(UTOBcastChunk.DEFAULT_NAME);
    chunk.setUTOBcastMessageType(UTOBcastChunk.ACK);

    // ... to leader
    try
    {
      chunk.setProcessTo(processMembershipItf.getLeader());
    }
    catch (InterruptedException e1)
    {
      throw new InterruptedPushException(e1);
    }
    try
    {
      // We do not clone message
      messageManagerItf.duplicateMessage(message, false);
      if (logger.isLoggable(BasicLevel.INFO))
      {
        logger.log(BasicLevel.INFO, "Send ACK message " + message
            + " to leader");
      }
      outPushItf.push(message, null);
    }
    catch (IOPushException e)
    {
      // We detect that the leader has crashed
      // TODO
      e.printStackTrace();
    }

    // ... to backup
    try
    {
      chunk.setProcessTo(processMembershipItf.getBackup());
    }
    catch (InterruptedException e1)
    {
      throw new InterruptedPushException(e1);
    }
    try
    {
      // We do not duplicate messages
      if (logger.isLoggable(BasicLevel.INFO))
      {
        logger.log(BasicLevel.INFO, "Send ACK message " + message
            + " to backup");
      }
      outPushItf.push(message, null);
    }
    catch (IOPushException e)
    {
      // We detect that the backup has crashed
      // TODO
      e.printStackTrace();
    }

  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(java.lang.String,
   *      java.lang.Object)
   */
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(DELIVERY_ITF_NAME))
    {
      deliveryItf = (Push) serverItf;
    }
    else if (clientItfName.equals(TO_BE_DELIVERED_IN_ITF_NAME))
    {
      toBeDeliveredInItf = (Push) serverItf;
    }
    else if (clientItfName.equals(TO_BE_DELIVERED_OUT_ITF_NAME))
    {
      toBeDeliveredOutItf = (Pull) serverItf;
    }
    else if (clientItfName.equals(Push.OUT_PUSH_ITF_NAME))
    {
      outPushItf = (Push) serverItf;
    }
    else if (clientItfName.equals(ProcessMembership.ITF_NAME))
    {
      processMembershipItf = (ProcessMembership) serverItf;
    }
    else if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{DELIVERY_ITF_NAME, TO_BE_DELIVERED_IN_ITF_NAME,
        TO_BE_DELIVERED_OUT_ITF_NAME, Push.OUT_PUSH_ITF_NAME,
        ProcessMembership.ITF_NAME, MessageManager.ITF_NAME};
  }

}