/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): 
 */

package org.objectweb.dream.pump;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.Pull;
import org.objectweb.dream.PullException;
import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.control.activity.ActiveComponentAttributeController;
import org.objectweb.dream.control.activity.Util;
import org.objectweb.dream.control.activity.task.AbstractTask;
import org.objectweb.dream.control.activity.task.Task;
import org.objectweb.dream.control.activity.task.TaskController;
import org.objectweb.dream.control.activity.task.thread.ThreadPoolController;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.synchro.Mutex;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.julia.control.lifecycle.ChainedIllegalLifeCycleException;

/**
 * Synchronized implementation of a Pull/Push pump. This component has an
 * optional {@link Mutex }client interface that it uses to synchronize the
 * execution of its run method with other components.
 */
public class PumpSynchronizedImpl extends AbstractComponent
    implements
      PumpAttributeController,
      ActiveComponentAttributeController
{

  /** The server interface to which this component is bound. */
  protected Push    outPushItf;

  /** The server interface to which this component is bound. */
  protected Pull    inPullItf;

  /** The server interface to which this component is bound. */
  protected Mutex   mutexItf;

  protected Task    pumpTask               = new PumpTask();

  protected boolean pushNullPolicy         = false;

  protected boolean usePool                = false;

  protected int     initialNumberOfThreads = 1;

  protected int     initialCapacity        = 1;

  // ---------------------------------------------------------------------------
  // Implementation of PumpAttributeController interface
  // ---------------------------------------------------------------------------

  /**
   * @see PumpAttributeController#getPushNullPolicy()
   */
  public boolean getPushNullPolicy()
  {
    return pushNullPolicy;
  }

  /**
   * @see PumpAttributeController#setPushNullPolicy(boolean)
   */
  public void setPushNullPolicy(boolean pushNull)
  {
    pushNullPolicy = pushNull;
  }

  // ---------------------------------------------------------------------------
  // Implementation of PumpTask class
  // ---------------------------------------------------------------------------

  class PumpTask extends AbstractTask
  {

    /**
     * Default constructor
     */
    public PumpTask()
    {
      super("PumpTask");
    }

    /**
     * @see Task#execute(Object)
     */
    public Object execute(Object hints) throws InterruptedException
    {
      if (mutexItf != null)
      {
        mutexItf.lock();
      }
      try
      {
        Message msg = inPullItf.pull(null);
        if (msg != null || pushNullPolicy)
        {
          outPushItf.push(msg, null);
        }
      }
      catch (PullException e)
      {
        throw new InterruptedException(e.getLocalizedMessage());
      }
      catch (PushException e)
      {
        throw new InterruptedException(e.getLocalizedMessage());
      }
      finally
      {
        if (mutexItf != null)
        {
          mutexItf.unlock();
        }
      }
      return EXECUTE_AGAIN;
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of the ActiveComponentAttributeController interface.
  // ---------------------------------------------------------------------------

  /**
   * @see ActiveComponentAttributeController#setUsePool(boolean)
   */
  public void setUsePool(boolean usePool)
  {
    this.usePool = usePool;
  }

  /**
   * @see ActiveComponentAttributeController#getUsePool()
   */
  public boolean getUsePool()
  {
    return usePool;
  }

  /**
   * @see ActiveComponentAttributeController#setInitialCapacity(int)
   */
  public void setInitialCapacity(int i)
  {
    this.initialCapacity = i;
  }

  /**
   * @see ActiveComponentAttributeController#getInitialCapacity()
   */
  public int getInitialCapacity()
  {
    return initialCapacity;
  }

  /**
   * @see ActiveComponentAttributeController#setInitialNumberOfThreads(int)
   */
  public void setInitialNumberOfThreads(int i)
  {
    this.initialNumberOfThreads = i;
  }

  /**
   * @see ActiveComponentAttributeController#getInitialNumberOfThreads()
   */
  public int getInitialNumberOfThreads()
  {
    return initialNumberOfThreads;
  }

  // -------------------------------------------------------------------------
  // Implementation of the LifeCycleController interface.
  // -------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.LifeCycleController#startFc()
   */
  public void startFc() throws IllegalLifeCycleException
  {
    super.startFc();
    if (usePool)
    {
      try
      {
        TaskController tc = (TaskController) weaveableC
            .getFcInterface("task-controller");
        ThreadPoolController threadPoolController = (ThreadPoolController) tc
            .getTaskControl(pumpTask);
        threadPoolController.addThreads(initialNumberOfThreads);
      }
      catch (Exception e)
      {
        throw new ChainedIllegalLifeCycleException(e, null,
            "An error occured while retrieving task controller interface");
      }
    }
  }

  /**
   * @see AbstractComponent#beforeFirstStart(Component)
   */
  protected void beforeFirstStart(Component componentItf)
      throws IllegalLifeCycleException
  {
    try
    {
      if (usePool)
      {
        Map m = new HashMap();
        m.put("thread", "pool");
        m.put("threadPool.capacity", new Integer(initialCapacity));
        Util.addTask(componentItf, pumpTask, m);
        // Note that threads cannot be added since the tasks is not started (so
        // it does not have any controller)
        // Threads are added in the startFc method.
      }
      else
      {
        Util.addTask(componentItf, pumpTask, Collections.EMPTY_MAP);
      }
    }
    catch (Exception e)
    {
      throw new IllegalLifeCycleException("Can't add task");
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{Push.OUT_PUSH_ITF_NAME, Pull.IN_PULL_ITF_NAME,
        Mutex.MUTEX_ITF_NAME};
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public synchronized void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(Push.OUT_PUSH_ITF_NAME))
    {
      outPushItf = (Push) serverItf;
    }
    else if (clientItfName.equals(Pull.IN_PULL_ITF_NAME))
    {
      inPullItf = (Pull) serverItf;
    }
    else if (clientItfName.equals(Mutex.MUTEX_ITF_NAME))
    {
      mutexItf = (Mutex) serverItf;
    }
  }

}