/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): 
 */

package org.objectweb.dream.queue;

import java.util.Map;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.InterruptedPullException;
import org.objectweb.dream.Pull;
import org.objectweb.dream.PullException;
import org.objectweb.dream.message.Message;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;

/**
 * Abstract implementation of a Pull outgoing handler. It implements overflow
 * policies as defined in {@link PullQueueAttributeController}. Moreover, it
 * implements the <code>pull</code> method: it checks for available message
 * and then applies appropriate policy.
 * <p>
 * Pull incoming handler developers can inherit this class. They must implement
 * two methods:
 * <ul>
 * <li><code>hasAvailableMessage()</code> returns a boolean indicating
 * whether a message is available for removal.</li>
 * <li><code>doPull()</code> removes a message from the buffer.</li>
 * </ul>
 */
public abstract class AbstractPullOutgoingHandlerImpl extends AbstractComponent
    implements
      Pull,
      PullQueueAttributeController
{

  protected boolean               blockingPull = true;

  // Client interfaces
  protected Buffer                bufferItf;
  protected BufferRemoveFirstLast bufferRemoveFirstLastItf;

  // ---------------------------------------------------------------------------
  // Implementation of the Pull interface.
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.Pull#pull(java.util.Map)
   */
  public synchronized Message pull(Map context) throws PullException
  {
    if (!hasAvailableMessage() && !blockingPull)
    {
      return null;
    }
    try
    {
      return doPull();
    }
    catch (InterruptedException e)
    {
      throw new InterruptedPullException(
          "Interrupted while waiting for an available message", e);
    }
  }

  // ---------------------------------------------------------------------------
  // Methods that must be implemented by concrete super classes.
  // ---------------------------------------------------------------------------

  /**
   * Checks whether there is an available message.
   * 
   * @return <code>true</code> if there is an available message.
   */
  protected abstract boolean hasAvailableMessage();

  /**
   * Pulls a message from the buffer (i.e. get or remove). This method should
   * not check if there is a message available provided it has already been done.
   * 
   * @return a message.
   * @throws InterruptedException if it is interrupted while waiting an
   *           available message.
   */
  protected abstract Message doPull() throws InterruptedException;

  // ---------------------------------------------------------------------------
  // Implementation of the PullQueueAttributeController interface.
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.queue.PullQueueAttributeController#setBlockingPull(boolean)
   */
  public synchronized void setBlockingPull(boolean blockingPull)
  {
    this.blockingPull = blockingPull;
  }

  /**
   * @see org.objectweb.dream.queue.PullQueueAttributeController#getBlockingPull()
   */
  public boolean getBlockingPull()
  {
    return blockingPull;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface.
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public synchronized void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(Buffer.ITF_NAME))
    {
      bufferItf = (Buffer) serverItf;
    }
    else if (clientItfName.equals(BufferRemoveFirstLast.ITF_NAME))
    {
      bufferRemoveFirstLastItf = (BufferRemoveFirstLast) serverItf;
    }
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{Buffer.ITF_NAME, BufferRemoveFirstLast.ITF_NAME};
  }

}