/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): 
 */

package org.objectweb.dream.queue;

import org.objectweb.dream.message.Message;

/**
 * This interface defines methods that must be implemented by a buffer. Buffers
 * are used to store messages. It is possible to
 * <ul>
 * <li>add and remove messages</li>
 * <li>get an indicator on stored messages</li>
 * <li>get an indicator on the available space in the buffer</li>
 * <li>get an indicator on messages available for removal</li>
 * </ul>
 * <p>
 * <i>Note 1: </i> all the messages that are stored in a buffer are not always
 * available for removal (e.g. when the buffer implements an ordering policy).
 * <p>
 * <i>Note 2: </i> the different indicators (stored messages, available space,
 * and available messages) have no predetermined semantics: the indicator can be
 * a number of messages, a number of bytes, etc.
 */
public interface Buffer
{
  /** The commonly used name to refer to this interface. */
  String ITF_NAME = "buffer";

  /**
   * Returns an indicator on stored messages. The semantic of this indicator
   * depends on the buffer implementation. Typical examples of indicators are
   * the number of messages, the number of bytes, etc.
   * 
   * @return an indicator on stored messages.
   */
  int storedMessagesIndicator();

  /**
   * Returns an indicator on available messages. The semantic of this indicator
   * depends on the buffer implementation. Typical examples of indicators are
   * the number of messages, the number of bytes, etc.
   * 
   * @return an indicator on available messages.
   */
  int availableMessagesIndicator();

  /**
   * Returns an indicator on available space. The semantic of this indicator
   * depends on the buffer implementation. Typical examples of indicators are
   * the number of messages, the number of bytes, etc. <i>Note: </i> when the
   * buffer has no maximum storage capabilities, this method must return
   * <code>Integer.MAX_VALUE</code>.
   * 
   * @return an indicator on available space.
   */
  int availableSpaceIndicator();

  /**
   * Adds the specified message to the buffer. This method may block if there is
   * not enough space available.
   * <p>
   * Buffers may place limitations on what elements may be added to this buffer.
   * In particular, some buffers require that messages contain particular
   * chunks.
   * 
   * @param message message to be addedto be appended to this list.
   * @throws InterruptedException if it is interrupted while adding the message.
   */
  void add(Message message) throws InterruptedException;

  /**
   * Removes a message from the buffer. This method blocks if
   * {@link #availableMessagesIndicator}returns <code>0</code>.
   * 
   * @return a message.
   * @throws InterruptedException if it is interrupted while removing a message.
   */
  Message remove() throws InterruptedException;

  /**
   * Gets a message from the buffer. This method blocks if
   * {@link #availableMessagesIndicator}returns <code>0</code>.
   * 
   * @return a message.
   * @throws InterruptedException if it is interrupted while removing a message.
   */
  Message get() throws InterruptedException;

}