/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): 
 */

package org.objectweb.dream.queue;

import java.util.Enumeration;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.message.manager.MessageManager;
import org.objectweb.dream.pool.ObjectPool;
import org.objectweb.dream.pool.Recyclable;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;

/**
 * The ListAddRemoveFirstLastFastImpl class is just a linked list that supports
 * very efficient insertion and deletion, as well as an Enumeration interface.
 * <p>
 * <i>Note: </i>None of the methods in this linked list are synchronized.
 * <p>
 * This class is freely inspired from code belonging to the SEDA project (see
 * http://sourceforge.net/projects/seda/).
 */
public class ListAddRemoveFirstLastFastImpl extends AbstractComponent
    implements
      List,
      ListAddFirstLast,
      ListRemoveFirstLast
{

  private int          numInList;
  private Element      first;
  private Element      last;

  /** The client interface used to access the pool of {@link Element}instances. */
  protected ObjectPool objectPoolItf;

  //---------------------------------------------------------------------------
  // Constructor.
  // ---------------------------------------------------------------------------

  /**
   * Allocates a brand new ListAddRemoveFirstLastFastImpl.
   */
  public ListAddRemoveFirstLastFastImpl()
  {
    numInList = 0;
    first = null;
    last = null;
  }

  //---------------------------------------------------------------------------
  // Implementation of the List interface.
  // ---------------------------------------------------------------------------

  /**
   * @see List#add(Object)
   */
  public void add(Object o)
  {
    addLast(o);

  }

  /**
   * @see List#remove()
   */
  public Object remove()
  {
    return removeFirst();
  }

  /**
   * @see List#isEmpty()
   */
  public boolean isEmpty()
  {
    return numInList == 0;
  }
  
  //---------------------------------------------------------------------------
  // Implementation of the ListAddFirstLast interface.
  // ---------------------------------------------------------------------------

  /**
   * @see ListAddFirstLast#addLast(Object)
   */
  public void addLast(Object o)
  {
    Element lle = (Element) objectPoolItf.newInstance();
    lle.obj = o;

    if (first == null)
    {
      first = lle;
      last = lle;
    }
    else
    {
      last.next = lle;
      lle.prev = last;
      last = lle;
    }
    numInList++;
  }

  /**
   * @see ListAddFirstLast#addFirst(Object)
   */
  public void addFirst(Object o)
  {
    Element lle = (Element) objectPoolItf.newInstance();
    lle.obj = o;

    if (first == null)
    {
      first = lle;
      last = lle;
    }
    else
    {
      first.prev = lle;
      lle.next = first;
      first = lle;
    }
    numInList++;
  }

  //---------------------------------------------------------------------------
  // Implementation of the ListRemoveFirstLast interface.
  // ---------------------------------------------------------------------------

  /**
   * @see ListRemoveFirstLast#getLast()
   */
  public Object getLast()
  {
    if (last == null)
    {
      return null;
    }
    return last.obj;
  }

  /**
   * @see ListRemoveFirstLast#removeLast()
   */
  public Object removeLast()
  {
    Element retme = null;
    Object retobj = null;

    if (last == null)
    {
      return null;
    }
    retme = last;

    if (first == last)
    {
      first = null;
      last = null;
    }
    else
    {
      last = last.prev;
      last.next = null;
    }

    numInList--;
    retobj = retme.obj;
    objectPoolItf.recycleInstance(retme);
    return retobj;
  }

  /**
   * @see ListRemoveFirstLast#getFirst()
   */
  public Object getFirst()
  {
    if (first == null)
    {
      return null;
    }
    return first.obj;
  }

  /**
   * @see ListRemoveFirstLast#removeFirst()
   */
  public Object removeFirst()
  {
    Element retme = null;
    Object retobj = null;

    if (first == null)
      return null;
    retme = first;

    if (first == last)
    {
      first = null;
      last = null;
    }
    else
    {
      first = first.next;
      first.prev = null;
    }

    numInList--;
    retobj = retme.obj;
    objectPoolItf.recycleInstance(retme);
    return retobj;
  }

  //---------------------------------------------------------------------------
  // Other methods.
  // ---------------------------------------------------------------------------

  /**
   * Returns the first object that is "equal" to the given object, based on the
   * response of the Object.equals() method.
   * 
   * @param known the object to which an element must be equal.
   * @return the matching object, or null if there is nothing that matches.
   */
  public Object getItem(Object known)
  {
    Element tmp;

    tmp = first;
    while (tmp != null)
    {
      if (known.equals(tmp.obj))
      {
        return tmp.obj;
      }
      tmp = tmp.next;
    }
    return null;
  }

  /**
   * Removes the first object that is "equal" to the given object, based on the
   * response of the Object.equals() method.
   * 
   * @param known the object to which an element must be equal.
   * @return the match, or null if there is nothing that matches.
   */
  public Object removeItem(Object known)
  {
    Element tmp;

    tmp = first;
    while (tmp != null)
    {
      if (known.equals(tmp.obj))
      {

        // cut it out and return it
        if (tmp == first)
        {
          return this.removeFirst();
        }
        else if (tmp == last)
        {
          return this.removeLast();
        }
        else
        {
          // somewhere in middle
          Object retobj;

          (tmp.prev).next = tmp.next;
          (tmp.next).prev = tmp.prev;
          numInList--;
          retobj = tmp.obj;
          objectPoolItf.recycleInstance(tmp);
          return retobj;
        }
      }
      tmp = tmp.next;
    }
    return null;
  }

  /**
   * Returns a java.util.Enumeration enumeration over the elements of the linked
   * list. If you modify the list while enumerating over it, you get what you
   * deserve (i.e. undefined behaviour).
   * 
   * @return the enumeration over the elements
   * @see Enumeration
   */
  public Enumeration elements()
  {
    return new FastLinkedListEnumeration();
  }

  //---------------------------------------------------------------------------
  // Overriden methods.
  // ---------------------------------------------------------------------------

  /**
   * @see Object#toString()
   */
  public String toString()
  {
    return toString("");
  }

  private String toString(String prefix)
  {
    String pre = prefix;
    if (pre == null)
      pre = "";

    String ret = pre + "ListAddRemoveFirstLastFastImpl: length=" + numInList
        + "\n";

    Element current = first;
    while (current != null)
    {
      if (current.obj == null)
      {
        ret += pre + "  (null)\n";
      }

      else
      {
        if (current.obj instanceof ListAddRemoveFirstLastFastImpl)
        {
          ret += pre
              + "LINKED LIST\n"
              + ((ListAddRemoveFirstLastFastImpl) current.obj).toString(pre
                  + "  ");
        }
        else
        {
          ret += pre + "  " + current.obj.toString() + "\n";
        }
      }

      current = current.next;
    }

    return ret;
  }

  // ---------------------------------------------------------------------------
  // Inner classes.
  // ---------------------------------------------------------------------------

  /**
   * This class represents elements of a linked list. Each element contains a
   * reference to its predecessor (<code>previous</code>), its successor (
   * <code>next</code>), a message, and a sequence number.
   */
  public static class Element implements Recyclable
  {
    Object  obj;
    Element prev;
    Element next;

    /**
     * @see org.objectweb.dream.pool.Recyclable#recycle()
     */
    public void recycle()
    {
      // Do nothing because code does not require anything.
    }

  }

  // ---------------------------------------------------------------------------
  // Inner classes implementing the Enumeration interface.
  // ---------------------------------------------------------------------------

  /**
   * A FastLinkedListEnumeration is a java.util.Enumeration over the
   * ListAddRemoveFirstLastFastImpl elements.
   * 
   * @see Enumeration
   */
  public class FastLinkedListEnumeration implements Enumeration
  {
    private Element curElement;

    /**
     * Constructor.
     */
    public FastLinkedListEnumeration()
    {
      curElement = ListAddRemoveFirstLastFastImpl.this.first;
    }

    /**
     * @see Enumeration#hasMoreElements()
     */
    public boolean hasMoreElements()
    {
      if (curElement == null)
        return false;
      return true;
    }

    /**
     * @see Enumeration#nextElement()
     */
    public Object nextElement() throws java.util.NoSuchElementException
    {
      Object retme;

      if (curElement == null)
        throw new java.util.NoSuchElementException();
      retme = curElement.obj;
      curElement = curElement.next;
      return retme;
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface.
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public synchronized void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(ObjectPool.ITF_NAME))
    {
      objectPoolItf = (ObjectPool) serverItf;
    }
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{MessageManager.ITF_NAME, ObjectPool.ITF_NAME};
  }

  // test code
  private static void printTime(long long1, long long3, int int5)
  {
    long long6 = long3 - long1;
    double double8 = (double) int5 / (double) long6;
    double double10 = double8 * 1000.0;

    System.out.println(int5 + " iterations in " + long6 + " milliseconds = "
        + double10 + " iterations per second");
  }

}

