/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): Vivien Quema
 */

package org.objectweb.dream.router;

import java.util.HashMap;
import java.util.Map;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.manager.MessageManager;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Abstract implementation of a basic Push/Push router. A Push/Push router
 * routes messages received on its Push input to one of its Push outputs. A
 * Push/Push router has also a <code>defaultOutPushItf</code> that can be
 * bound to route messages for which no routes are found. <br>
 * Outputs are stored in the <code>outPushMap</code> Map. The output
 * corresponding to an incoming message is retrieved using the abstract
 * {@link #getOutput(Message, Map) }method. To avoid unnecessary building of
 * routes the {@link AbstractRouterImpl#getOutput(Message, Map) }can test
 * the boolean <code>initialized</code> which is set to <code>false</code>
 * each time the router is bound to a new output.
 */
public abstract class AbstractRouterImpl extends AbstractComponent
    implements
      Push
{

  /** Map of out-push interfaces binded to this router */
  protected Map              outPushMap;

  /** The name of the default {@link Push}client interface. */
  public static final String DEFAULT_OUT_PUSH_ITF_NAME = "defaultOutPush";

  /**
   * Default Push interface to be used if no Push interface can be found in
   * <code>outPushTable</code>.
   */
  protected Push             defaultOutPushItf         = null;

  /** the message manager client interface of this component */
  protected MessageManager   messageManagerItf;

  /** <code>true</code> if the route table is initialized. */
  protected boolean          initialized;

  /** Default contructor */
  public AbstractRouterImpl()
  {
    outPushMap = new HashMap();
    initialized = false;
  }

  protected abstract Push getOutput(Message message, Map context)
      throws PushException;

  /**
   * Sends the incoming message to one of the router's output. The appropriate
   * output is retrieved using the {@link #getOutput(Message, Map) }method. If
   * this method returns <code>null</code>, the
   * <code>defaultOutPushItf</code> is used if it is bound.
   * 
   * @param message the pushed message
   * @param context a context propagated to the selected route.
   * @throws PushException if no route can be found, or thown by client push.
   * @see Push#push(Message, Map)
   */
  public void push(Message message, Map context) throws PushException
  {
    Push outPush = getOutput(message, context);
    if (outPush == null)
    {
      if (defaultOutPushItf == null)
      {
        logger.log(BasicLevel.ERROR, "No route for message " + message
            + ". The message will be dropped");
        messageManagerItf.deleteMessage(message);
        return;
      }
      defaultOutPushItf.push(message, context);
    }
    else
    {
      outPush.push(message, context);
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    String[] tab = new String[outPushMap.size() + 1];
    outPushMap.keySet().toArray(tab);
    tab[outPushMap.size()] = MessageManager.ITF_NAME;
    return tab;
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public synchronized void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.startsWith(OUT_PUSH_ITF_NAME))
    {
      outPushMap.put(clientItfName, serverItf);
      initialized = false;
    }
    else if (clientItfName.equals(DEFAULT_OUT_PUSH_ITF_NAME))
    {
      defaultOutPushItf = (Push) serverItf;
      initialized = false;
    }
    else if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#unbindFc(String)
   */
  public synchronized void unbindFc(String clientItfName)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.unbindFc(clientItfName);
    if (clientItfName.startsWith(OUT_PUSH_ITF_NAME))
    {
      outPushMap.remove(clientItfName);
      initialized = false;
    }
    else if (clientItfName.equals(DEFAULT_OUT_PUSH_ITF_NAME))
    {
      defaultOutPushItf = null;
      initialized = false;
    }
  }
}